import pytest
from hypothesis import given, settings, reproduce_failure, unlimited, HealthCheck, seed
from hypothesis.extra.pandas import data_frames, columns, range_indexes, column, indexes
from hypothesis.extra.numpy import arrays
import hypothesis.strategies as st

from itertools import product
import tempfile
import subprocess
from io import StringIO

from pyranges import PyRanges

# using assert df equal, because we want to consider output from bedtools and
# pyranges equal even if they have different sort order
from tests.helpers import assert_df_equal

import numpy as np

from os import environ

if environ.get("TRAVIS"):
    max_examples = 100
    deadline = None
else:
    max_examples = 100
    deadline = None


def mysort(tp):

    if tp[1] == tp[2]:
        tp = (tp[0], tp[1], tp[2] + 1, tp[3])

    key = [-1, tp[1], tp[2], int(1e10)]

    return [x for _, x in sorted(zip(key, tp))]

chromosomes = st.sampled_from(["chr{}".format(str(e)) for e in list(range(1, 23)) + "X Y M".split()])
chromosomes_small = st.sampled_from(["chr1"])
cs = st.one_of(chromosomes, chromosomes_small)

positions = st.integers(min_value=0, max_value=int(1e7))
lengths = st.integers(min_value=1, max_value=int(1e7))
small_lengths = st.integers(min_value=1, max_value=int(1e4))
strands = st.sampled_from("+ -".split())
names = st.text("abcdefghijklmnopqrstuvxyz", min_size=1)
scores = st.integers(min_value=0, max_value=256)


# dfs = data_frames(columns=columns("Chromosome Start End Strand".split(),
#                                   dtype=int), rows=st.tuples(chromosomes, positions, positions,
#                                                              strands).map(mysort))

df_minsize = 1
nonempty_dfs = data_frames(index=indexes(dtype=np.int64, min_size=df_minsize, unique=True),
                           columns=columns("Chromosome Start End Strand".split(), dtype=int),
                           rows=st.tuples(chromosomes, positions, positions, strands).map(mysort))


better_df_minsize = 1
better_dfs = data_frames(index=indexes(dtype=np.int64, min_size=better_df_minsize, unique=True),
                         columns=[column("Chromosome", chromosomes),
                                  column("Start", elements=positions),
                                  column("End", elements=lengths),
                                  column("Strand", strands)])

better_dfs_min = data_frames(index=indexes(dtype=np.int64, min_size=better_df_minsize, unique=True),
                             columns=[column("Chromosome", cs),
                                      column("Start", elements=lengths),
                                      column("End", elements=small_lengths),
                                      # column("Name", elements=names),
                                      # column("Score", elements=scores),
                                      column("Strand", strands)])

better_dfs_min_single_chromosome = data_frames(index=indexes(dtype=np.int64, min_size=better_df_minsize, unique=True),
                                               columns=[column("Chromosome", chromosomes_small),
                                                        column("Start", elements=lengths),
                                                        column("End", elements=small_lengths),
                                                        column("Name", elements=names),
                                                        column("Score", elements=scores),
                                                        column("Strand", strands)])

better_dfs_min_nostrand = data_frames(index=indexes(dtype=np.int64, min_size=better_df_minsize, unique=True),
                             columns=[column("Chromosome", chromosomes_small),
                                      column("Start", elements=small_lengths),
                                      column("End", elements=small_lengths)])


large_df_minsize = 5
large_dfs = data_frames(index=indexes(dtype=np.int64, min_size=large_df_minsize, unique=True),
                           columns=columns("Chromosome Start End Name Score Strand".split(), dtype=int),
                           rows=st.tuples(chromosomes, positions, positions, names, small_lengths, strands).map(mysort))

runlengths = data_frames(index=indexes(dtype=np.int64, min_size=df_minsize, unique=True),
                         columns=[column("Runs", st.integers(min_value=1, max_value=int(1e7))),
                                  # must have a min/max on floats because R S4vectors translates too big ones into inf.
                                  # which is unequal to eg -1.79769e+308 so the tests fail
                                  column("Values", st.integers(min_value=-int(1e7), max_value=int(1e7)))])

# runlengths_same_length_floats = data_frames(index=range_indexes(min_size=df_minsize),
#                                             columns=[column("Runs", st.integers(min_value=1, max_value=int(1e7))),
#                                                      column("Values", st.floats(min_value=0.001, max_value=int(1e7))),
#                                                      column("Values2", st.floats(min_value=0.001, max_value=int(1e7)))])


runlengths_same_length_integers = data_frames(index=indexes(dtype=np.int64, min_size=df_minsize, unique=True),
                                              columns=[column("Runs", st.integers(min_value=1, max_value=int(1e4))),
                                              column("Values", st.integers(min_value=1, max_value=int(1e4))),
                                              column("Values2", st.integers(min_value=1, max_value=int(1e4)))])


# runs = arrays(st.integers(min_value=1, max_value=int(1e5)), shape=)
# values = arrays(dtype=np.float)
@st.composite
def dfs_min(draw):
    df = draw(better_dfs_min)
    df.loc[:, "End"] += df.Start

    df.insert(3, "Name", "a")
    df.insert(4, "Score", 0)
    gr = PyRanges(df)

    # do not sort like this, use pyranges sort
    # np.random.seed(draw(st.integers(min_value=0, max_value=int(1e6))))
    # gr.df = df.reindex(np.random.permutation(df.index.values))

    return gr


@st.composite
def dfs_min_single_chromosome(draw):
    df = draw(better_dfs_min_single_chromosome)
    df.loc[:, "End"] += df.Start
    df.insert(3, "Name", "a")
    df.insert(4, "Score", 0)

    return df


import pyranges as pr
import pandas as pd
from pyrle import Rle
import numpy as np


strandedness = [False, "same", "opposite"]
methods = ["join", "intersection", "overlap"]

# strandedness = strandedness[1:2]
# methods = methods[1:2]


coverage_cmd = "Rscript --vanilla tests/compute_coverage.R {} {}"

@pytest.mark.r
@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@given(df=dfs_min_single_chromosome())
def test_coverage(df):

    print("---" * 10)
    p = pr.PyRanges(df)
    print("pyranges\n", p)

    c = p.coverage()["chr1"]

    result_df = None
    with tempfile.TemporaryDirectory() as temp_dir:
        f1 = "{}/f1.txt".format(temp_dir)
        outfile = "{}/result.txt".format(temp_dir)
        R_df = df
        R_df.End = R_df.End - 1
        R_df.to_csv(f1, sep="\t", index=False)

        cmd = coverage_cmd.format(f1, outfile) + " 2>/dev/null"

        subprocess.check_output(cmd, shell=True, executable="/bin/bash").decode()

        result = pd.read_table(outfile)[["Runs.value", "Values.value"]]
        result.columns = "Runs Values".split()
        result = pd.concat([pd.DataFrame(index=[0], data={"Runs": 1, "Values": 0}), result], ignore_index=True)
        s4vectors_result = Rle(result.Runs, result.Values)

    print("pyranges result\n", c)
    print("s4vectors result\n", s4vectors_result)
    print(str(c == s4vectors_result) + " " * 10, c == s4vectors_result)

    assert np.all(np.equal(c.runs, s4vectors_result.runs))
    assert np.all(np.equal(c.values, s4vectors_result.values))


rle_operations = "+ - / *".split()
# rle_operations = ["-"]

rle_operation_cmd = "Rscript --vanilla tests/compute_Rle.R {} {} '{}' {}"

@pytest.mark.r
@given(runlengths=runlengths, runlengths2=runlengths)
@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@pytest.mark.parametrize("operation", rle_operations)
def test_rle(runlengths, runlengths2, operation):

    # It is only compared against bioc with integers because float equality is hard,
    # for both libraries, sometimes end up with slightly different runlengths
    # [257.492105948544, 257.492075430654] gives
    # pyranges result
    #  +--------+---------+---------+---------+
    # | Runs   |       1 |       1 |       1 |
    # |--------+---------+---------+---------|
    # | Values | 257.492 | 257.492 | 257.492 |
    # +--------+---------+---------+---------+
    # Rle of length 3 containing 3 elements
    # s4vectors result
    #  +--------+---------+---------+
    # | Runs   |       2 |       1 |
    # |--------+---------+---------|
    # | Values | 257.492 | 257.492 |
    # +--------+---------+---------+
    # Rle of length 3 containing 2 elements

    pyop = {"+": "__add__", "-": "__sub__", "*": "__mul__", "/": "__truediv__"}[operation]

    print("runlengths", runlengths)
    print("runlengths2", runlengths2)

    r = Rle(runlengths.Runs, runlengths.Values)
    r2 = Rle(runlengths2.Runs, runlengths2.Values)

    print("r\n", r)
    print("r2\n", r2)

    m = getattr(r, pyop)
    result_pyranges = m(r2)

    print("pyranges result\n", result_pyranges)
    # f1 = "f1.txt"
    # f2 = "f2.txt"
    # runlengths.to_csv(f1, sep="\t", index=False)
    # runlengths2.to_csv(f2, sep="\t", index=False)
    result_df = None
    with tempfile.TemporaryDirectory() as temp_dir:
        f1 = "{}/f1.txt".format(temp_dir)
        f2 = "{}/f2.txt".format(temp_dir)
        outfile = "{}/result.txt".format(temp_dir)
        runlengths.to_csv(f1, sep="\t", index=False)
        runlengths2.to_csv(f2, sep="\t", index=False)

        cmd = rle_operation_cmd.format(f1, f2, operation, outfile) + " 2>/dev/null"

        subprocess.check_output(cmd, shell=True, executable="/bin/bash").decode()

        result = pd.read_table(outfile)
        s4vectors_result = Rle(result.Runs, result.Values)

    print("pyranges result\n", result_pyranges)
    print("s4vectors result\n", s4vectors_result)

    assert np.allclose(result_pyranges.runs, s4vectors_result.runs, equal_nan=False)
    assert np.allclose(result_pyranges.values, s4vectors_result.values, equal_nan=True)

set_intersection_command = "bedtools intersect -a <(sort -k1,1 -k2,2n {} | bedtools merge -i -) -b <(sort -k1,1 -k2,2n {} | bedtools merge -i -)"


def name_score(df, df2):
    df.insert(3, "Name", "a")
    df.insert(4, "Score", 0)

    df2.insert(3, "Name", "b")
    df2.insert(4, "Score", 0)

    return df, df2

merge_command = "bedtools merge -o first -c 6 {} -i <(sort -k1,1 -k2,2n {})"
@pytest.mark.bedtools
@pytest.mark.parametrize("strand", [True, False])
@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@given(gr=dfs_min())
def test_cluster(gr, strand):

    bedtools_strand = {True: "-s", False: ""}[strand]

    print(gr)

    result_df = None
    with tempfile.TemporaryDirectory() as temp_dir:
        f1 = "{}/f1.bed".format(temp_dir)
        gr.df.to_csv(f1, sep="\t", header=False, index=False)

        cmd = merge_command.format(bedtools_strand, f1)
        print(cmd)

        result = subprocess.check_output(cmd, shell=True, executable="/bin/bash").decode()

        if not strand:
            bedtools_df = pd.read_table(StringIO(result), header=None, squeeze=True, usecols=[0, 1, 2], names="Chromosome Start End".split(), dtype={"Chromosome": "category"})
        else:
            bedtools_df = pd.read_table(StringIO(result), header=None, squeeze=True, names="Chromosome Start End Strand".split(), dtype={"Chromosome": "category"})

    print("bedtools_df\n", bedtools_df)
    result = gr.cluster(strand=strand)
    print("result\n", result.df)

    if not bedtools_df.empty:
        assert_df_equal(result.df, bedtools_df)
    else:
        assert bedtools_df.empty == result.df.empty


@pytest.mark.bedtools
@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())
def test_set_intersection(gr, gr2):

    print(gr)
    print(gr2)

    result_df = None
    with tempfile.TemporaryDirectory() as temp_dir:
        f1 = "{}/f1.bed".format(temp_dir)
        f2 = "{}/f2.bed".format(temp_dir)
        gr.df.to_csv(f1, sep="\t", header=False, index=False)
        gr2.df.to_csv(f2, sep="\t", header=False, index=False)

        cmd = set_intersection_command.format(f1, f2)

        result = subprocess.check_output(cmd, shell=True, executable="/bin/bash").decode()

        bedtools_df = pd.read_table(StringIO(result), header=None, squeeze=True, names="Chromosome Start End".split())

    print("bedtools_df\n", bedtools_df)
    result = gr.set_intersection(gr2)
    print("result\n", result.df)

    if not bedtools_df.empty:
        assert_df_equal(result.df, bedtools_df)
    else:
        assert bedtools_df.empty == result.df.empty


overlap_command = "bedtools intersect -u {} -a {} -b {}"

@pytest.mark.bedtools
@pytest.mark.parametrize("strandedness", strandedness)
@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())
def test_overlap(gr, gr2, strandedness):

    bedtools_strand = {False: "", "same": "-s", "opposite": "-S"}[strandedness]

    print(gr.df)
    print(gr2.df)

    result_df = None
    with tempfile.TemporaryDirectory() as temp_dir:
        f1 = "{}/f1.bed".format(temp_dir)
        f2 = "{}/f2.bed".format(temp_dir)
        gr.df.to_csv(f1, sep="\t", header=False, index=False)
        gr2.df.to_csv(f2, sep="\t", header=False, index=False)

        cmd = overlap_command.format(bedtools_strand, f1, f2)

        result = subprocess.check_output(cmd, shell=True, executable="/bin/bash").decode()

        bedtools_df = pd.read_table(StringIO(result), header=None, squeeze=True, names="Chromosome Start End Name Score Strand".split())

    result = gr.overlap(gr2, strandedness=strandedness)
    # print("result\n", result.df)
    # print("bedtools_df\n", bedtools_df)

    if not bedtools_df.empty:
        assert_df_equal(result.df, bedtools_df)
    else:
        assert bedtools_df.empty == result.df.empty

intersection_command = "bedtools intersect {} -a {} -b {}"

@pytest.mark.bedtools
@pytest.mark.parametrize("strandedness", strandedness)
@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())
def test_intersection(gr, gr2, strandedness):

    bedtools_strand = {False: "", "same": "-s", "opposite": "-S"}[strandedness]

    print(gr.df)
    print(gr2.df)

    result_df = None
    with tempfile.TemporaryDirectory() as temp_dir:
        f1 = "{}/f1.bed".format(temp_dir)
        f2 = "{}/f2.bed".format(temp_dir)
        gr.df.to_csv(f1, sep="\t", header=False, index=False)
        gr2.df.to_csv(f2, sep="\t", header=False, index=False)

        cmd = intersection_command.format(bedtools_strand, f1, f2)

        result = subprocess.check_output(cmd, shell=True, executable="/bin/bash").decode()

        bedtools_df = pd.read_table(StringIO(result), header=None, squeeze=True, names="Chromosome Start End Name Score Strand".split())

    result = gr.intersection(gr2, strandedness=strandedness)
    print("result\n", result.df)
    print("bedtools_df\n", bedtools_df)

    if not bedtools_df.empty:
        assert_df_equal(result.df, bedtools_df)
    else:
        assert bedtools_df.empty == result.df.empty


set_union_command = "bedtools merge -i <(sort -k1,1 -k2,2n --merge <(sort -k1,1 -k2,2n {} | bedtools merge -i -) <(sort -k1,1 -k2,2n {} | bedtools merge -i -))"

@pytest.mark.bedtools
@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())
def test_set_union(gr, gr2):

    result_df = None
    with tempfile.TemporaryDirectory() as temp_dir:
        f1 = "{}/f1.bed".format(temp_dir)
        f2 = "{}/f2.bed".format(temp_dir)
        gr.df.to_csv(f1, sep="\t", header=False, index=False)
        gr2.df.to_csv(f2, sep="\t", header=False, index=False)

        cmd = set_union_command.format(f1, f2)

        result = subprocess.check_output(cmd, shell=True, executable="/bin/bash").decode()

        bedtools_df = pd.read_table(StringIO(result), header=None, squeeze=True, names="Chromosome Start End".split())

    result = gr.set_union(gr2)
    print("result\n", result)
    print("bedtools_df\n", PyRanges(bedtools_df))

    if not bedtools_df.empty:
        assert_df_equal(result.df, bedtools_df)
    else:
        assert bedtools_df.empty == result.df.empty

subtraction_command = "bedtools subtract {} -a {} -b {}"

@pytest.mark.bedtools
@pytest.mark.parametrize("strandedness", ["same"]) # , False
@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())
def test_subtraction(gr, gr2, strandedness):

    # print("gr\n", gr)
    # print("gr2\n", gr2)
    bedtools_strand = {False: "", "same": "-s", "opposite": "-S"}[strandedness]

    result_df = None
    with tempfile.TemporaryDirectory() as temp_dir:
        f1 = "{}/f1.bed".format(temp_dir)
        f2 = "{}/f2.bed".format(temp_dir)
        gr.df.to_csv(f1, sep="\t", header=False, index=False)
        gr2.df.to_csv(f2, sep="\t", header=False, index=False)

        cmd = subtraction_command.format(bedtools_strand, f1, f2)

        result = subprocess.check_output(cmd, shell=True, executable="/bin/bash").decode()

        bedtools_df = pd.read_table(StringIO(result), header=None, squeeze=True, names="Chromosome Start End Name Score Strand".split())

    result = gr.subtraction(gr2, strandedness=strandedness)
    print("result\n", result)
    print("bedtools_df\n", PyRanges(bedtools_df))

    if not bedtools_df.empty:
        assert_df_equal(result.df, bedtools_df)
    else:
        assert bedtools_df.empty == result.df.empty


how = [False, "nonoverlapping", "previous_nonoverlapping", "next_nonoverlapping", "next", "previous"]

rle_commute_how = ["__add__", "__mul__"]

@pytest.mark.parametrize("how", rle_commute_how)
@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())
def test_commutative_rles(gr, gr2, how):

    cv = gr.coverage(stranded=True)
    cv2 = gr2.coverage(stranded=True)

    method = getattr(cv, how)
    method2 = getattr(cv2, how)

    result = method(cv2)
    result2 = method2(cv)

    assert result == result2, "\n".join([str(e) for e in [cv, cv2, result, result2, "---" * 10]])

@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@given(df=runlengths_same_length_integers)
def test_inverse_div_mul_rles(df):

    """Testing with small integers, since small value floating points might lead to
mul then div not being equal to identity function because of float equality."""

    print(df)
    runlength = df.Runs.sum()

    cv = Rle(df.Runs.values, df.Values.values)

    newruns = np.random.permutation(df.Runs.values)
    print("newruns", newruns)
    cv2 = Rle(newruns, df.Values2.values)

    print("cv\n", cv)
    print("cv2\n", cv2)

    assert runlength == np.sum(cv.runs) and runlength == np.sum(cv2.runs)

    result = cv / cv2

    result2 = result * cv2

    print("result\n", result)
    print("result2\n", result2)

    assert np.all(np.equal(result2.runs, cv.runs))
    assert np.allclose(result2.values, cv.values)


@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@given(df=runlengths_same_length_integers)
def test_inverse_add_sub_rles(df):

    """Testing with small integers, since small value floating points might lead to
mul then div not being equal to identity function because of float equality."""

    cv = Rle(df.Runs.values, df.Values.values)

    cv2 = Rle(np.random.permutation(df.Runs.values), df.Values2.values)

    result = cv + cv2

    result2 = result - cv2

    assert np.all(np.equal(result2.runs, cv.runs))
    assert np.allclose(result2.values, cv.values)


nearest_command = "bedtools closest {} {} -t first -d -a <(sort -k1,1 -k2,2n {}) -b <(sort -k1,1 -k2,2n {})"
                    # "bedtools closest {} -t first -io -d -a <(sort -k1,1 -k2,2n {}) -b <(sort -k1,1 -k2,2n {})"]
nearest_hows = [None, "previous", "next"]
overlaps = [True, False]
strandedness = [False, "same", "opposite"]


@pytest.mark.bedtools
@pytest.mark.parametrize("nearest_how,overlap,strandedness", product(nearest_hows, overlaps, strandedness))
@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())
def test_nearest(gr, gr2, nearest_how, overlap, strandedness):

    print(gr.df)
    print(gr2.df)

    bedtools_strand = {False: "", "same": "-s", "opposite": "-S"}[strandedness]
    bedtools_overlap = {True: "", False: "-io"}[overlap]

    result_df = None
    with tempfile.TemporaryDirectory() as temp_dir:
        f1 = "{}/f1.bed".format(temp_dir)
        f2 = "{}/f2.bed".format(temp_dir)
        gr.df.to_csv(f1, sep="\t", header=False, index=False)
        gr2.df.to_csv(f2, sep="\t", header=False, index=False)

        cmd = nearest_command.format(bedtools_overlap, bedtools_strand, f1, f2)
        # print(cmd)
        result = subprocess.check_output(cmd, shell=True, executable="/bin/bash").decode()

        bedtools_df = pd.read_table(StringIO(result), header=None, squeeze=True, names="Chromosome Start End Name Score Strand Chromosome_b Start_b End_b Name_b Score_b Strand_b Distance".split())
        print("bedtools_df", bedtools_df)
        # raise
        if not bedtools_df.empty:
            bedtools_df = bedtools_df[bedtools_df.Distance != -1]["Chromosome Start End Strand Distance".split()].sort_values("Distance")

    result = gr.nearest(gr2, overlap=overlap, strandedness=strandedness)
    if not result.df.empty:

        print("pyranges_df", result.df)
        result_df = result.df.sort_values("Distance")["Chromosome Start End Strand Distance".split()]
        # print("bedtools_df", bedtools_df)
        # print("pyranges", pyranges_distances)

        assert_df_equal(result_df, bedtools_df)
    else:
        result_df = pd.DataFrame(columns="Chromosome Start End Strand Distance".split())
        assert bedtools_df.empty

    # if not result.df.empty:
    #     pyranges_distances = result_df.Distance.tolist()
    # else:
    #     pyranges_distances = []

    # print("bedtools", bedtools_distances)



jaccard_strandedness = [False, "same"]
jaccard_command = "bedtools jaccard {}  -a <(sort -k1,1 -k2,2n {}) -b <(sort -k1,1 -k2,2n {})"
@pytest.mark.bedtools
@pytest.mark.parametrize("strandedness", jaccard_strandedness)
@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())
def test_jaccard(gr, gr2, strandedness):

    print(gr.df)
    print(gr2.df)

    bedtools_strand = {False: "", "same": "-s", "opposite": "-S"}[strandedness]
    print(bedtools_strand)

    result_df = None
    with tempfile.TemporaryDirectory() as temp_dir:
        f1 = "{}/f1.bed".format(temp_dir)
        f2 = "{}/f2.bed".format(temp_dir)
        gr.df.to_csv(f1, sep="\t", header=False, index=False)
        gr2.df.to_csv(f2, sep="\t", header=False, index=False)

        cmd = jaccard_command.format(bedtools_strand, f1, f2)
        print(f1)
        print(f2)
        print(cmd)
        result = subprocess.check_output(cmd, shell=True, executable="/bin/bash").decode()

        bedtools_jaccard = float(result.split("\n")[1].split()[2])

    print("result bedtools", bedtools_jaccard)
    result = gr.jaccard(gr2, strandedness=strandedness)
    print("pyranges result", result)

    assert 0 <= result <=1
    # sicne bedtools is buggy for this func?
    # assert np.isclose(result, bedtools_jaccard)


join_command = "bedtools intersect {} -wo -a {} -b {}"
join_strandedness = [False, "opposite", "same"]

@pytest.mark.bedtools
@pytest.mark.parametrize("strandedness", join_strandedness)
@settings(max_examples=max_examples, deadline=deadline, timeout=unlimited, suppress_health_check=HealthCheck.all())
@given(gr=dfs_min(), gr2=dfs_min())
def test_join(gr, gr2, strandedness):

    print(gr.df)
    print(gr2.df)

    bedtools_strand = {False: "", "same": "-s", "opposite": "-S"}[strandedness]

    result_df = None
    with tempfile.TemporaryDirectory() as temp_dir:
        f1 = "{}/f1.bed".format(temp_dir)
        f2 = "{}/f2.bed".format(temp_dir)
        gr.df.to_csv(f1, sep="\t", header=False, index=False)
        gr2.df.to_csv(f2, sep="\t", header=False, index=False)

        cmd = join_command.format(bedtools_strand, f1, f2)
        print(cmd)
        result = subprocess.check_output(cmd, shell=True, executable="/bin/bash").decode()

        bedtools_df = pd.read_table(StringIO(result), header=None, squeeze=True, names="Chromosome Start End Name Score Strand Chromosome_b Start_b End_b Name_b Score_b Strand_b Overlap".split(), dtype={"Chromosome": "category", "Strand": "category"} )
        bedtools_df = bedtools_df.drop(["Overlap", "Chromosome_b"], 1)

    print("gr\n", gr)
    print("gr2\n", gr2)

    result = gr.join(gr2, strandedness=strandedness)

    print("result\n", result)
    print("bedtools\n", PyRanges(bedtools_df))

    if result.df.empty:
        assert bedtools_df.empty
    else:
        assert_df_equal(result.df, bedtools_df)




import pytest
from tests.helpers import assert_df_equal

from pyranges.pyranges import PyRanges
import pyranges as pr

import pandas as pd

from io import StringIO


@pytest.fixture
def expected_result_unstranded():

    c = """Chromosome Start End Name Score Strand Start_b End_b Name_b Score_b Strand_b Distance
0 chr1 3 6 h 0 + 6 7 f 0 - 1"""

    return PyRanges(pd.read_table(StringIO(c), sep=" "))


def test_nearest_bed_unstranded(f1, f2, expected_result_unstranded):

    result = f1.nearest(f2, strandedness=False, suffix="_b", how="next", overlap=False)


    assert_df_equal(result.df, expected_result_unstranded.df)


@pytest.fixture
def expected_result_opposite_stranded():

    c = """Chromosome Start End Name Score Strand Start_b End_b Name_b Score_b Strand_b Distance
0 chr1 9939 10138 HWI-ST216_313:3:2301:15791:16298 1 + 10149 10348 HWI-ST216:427:D29R1ACXX:2:2306:7654:33038 1 - 12
1 chr1 9953 10152 HWI-ST216_313:3:1305:6975:102491 1 + 19958 20157 HWI-ST216:427:D29R1ACXX:2:1313:6283:67310 1 - 9807
2 chr1 10024 10223 HWI-ST216_313:3:2201:5209:155139 1 + 19958 20157 HWI-ST216:427:D29R1ACXX:2:1313:6283:67310 1 - 9736
3 chr1 10246 10445 HWI-ST216_313:3:1207:4315:142177 1 + 19958 20157 HWI-ST216:427:D29R1ACXX:2:1313:6283:67310 1 - 9514
4 chr1 9916 10115 HWI-ST216_313:3:1203:10227:6568 1 - 10280 10479 HWI-ST216:427:D29R1ACXX:2:2103:15965:69480 1 + 166
5 chr1 9951 10150 HWI-ST216_313:3:2205:20086:33508 1 - 10280 10479 HWI-ST216:427:D29R1ACXX:2:2103:15965:69480 1 + 131
6 chr1 9978 10177 HWI-ST216_313:3:1204:5599:113305 1 - 10280 10479 HWI-ST216:427:D29R1ACXX:2:2103:15965:69480 1 + 104
7 chr1 10001 10200 HWI-ST216_313:3:1102:14019:151362 1 - 10280 10479 HWI-ST216:427:D29R1ACXX:2:2103:15965:69480 1 + 81
8 chr1 10127 10326 HWI-ST216_313:3:2207:7406:122346 1 - 16056 16255 HWI-ST216:427:D29R1ACXX:2:1102:7604:12113 1 + 5731
9 chr1 10241 10440 HWI-ST216_313:3:1302:4516:156396 1 - 16056 16255 HWI-ST216:427:D29R1ACXX:2:1102:7604:12113 1 + 5617"""

    gr = PyRanges(pd.read_table(StringIO(c), sep=" "))

    return gr


def test_nearest_bed_opposite_stranded(chip_10_plus_one, input_10, expected_result_opposite_stranded):

    result = chip_10_plus_one.nearest(input_10, strandedness="opposite", suffix="_b", how="next", overlap=False)

    print("result", result.df.to_csv(sep=" "))

    assert_df_equal(result.df, expected_result_opposite_stranded.df)


@pytest.fixture
def expected_result_same_stranded():

    c = """Chromosome Start End Name Score Strand Start_b End_b Name_b Score_b Strand_b Distance
0 chr1 9916 10115 HWI-ST216_313:3:1203:10227:6568 1 - 10149 10348 HWI-ST216:427:D29R1ACXX:2:2306:7654:33038 1 - 35
1 chr1 9939 10138 HWI-ST216_313:3:2301:15791:16298 1 + 10280 10479 HWI-ST216:427:D29R1ACXX:2:2103:15965:69480 1 + 143
2 chr1 9951 10150 HWI-ST216_313:3:2205:20086:33508 1 - 19958 20157 HWI-ST216:427:D29R1ACXX:2:1313:6283:67310 1 - 9809
3 chr1 9953 10152 HWI-ST216_313:3:1305:6975:102491 1 + 10280 10479 HWI-ST216:427:D29R1ACXX:2:2103:15965:69480 1 + 129
4 chr1 9978 10177 HWI-ST216_313:3:1204:5599:113305 1 - 19958 20157 HWI-ST216:427:D29R1ACXX:2:1313:6283:67310 1 - 9782
5 chr1 10001 10200 HWI-ST216_313:3:1102:14019:151362 1 - 19958 20157 HWI-ST216:427:D29R1ACXX:2:1313:6283:67310 1 - 9759
6 chr1 10024 10223 HWI-ST216_313:3:2201:5209:155139 1 + 10280 10479 HWI-ST216:427:D29R1ACXX:2:2103:15965:69480 1 + 58
7 chr1 10127 10326 HWI-ST216_313:3:2207:7406:122346 1 - 19958 20157 HWI-ST216:427:D29R1ACXX:2:1313:6283:67310 1 - 9633
8 chr1 10241 10440 HWI-ST216_313:3:1302:4516:156396 1 - 19958 20157 HWI-ST216:427:D29R1ACXX:2:1313:6283:67310 1 - 9519
9 chr1 10246 10445 HWI-ST216_313:3:1207:4315:142177 1 + 16056 16255 HWI-ST216:427:D29R1ACXX:2:1102:7604:12113 1 + 5612"""

    return PyRanges(pd.read_table(StringIO(c), sep=" "))


def test_nearest_bed_same_stranded(chip_10_plus_one, input_10, expected_result_same_stranded):

    result = chip_10_plus_one.nearest(input_10, strandedness="same", suffix="_b", how="next", overlap=False)

    print("result", result.df.to_csv(sep=" "))

    assert_df_equal(result.df, expected_result_same_stranded.df)



@pytest.fixture()
def expected_result_counterexample1():

    c = """chr1  1  2  +  2   3   +         1"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None,
                                  names="Chromosome Start End Strand Start_b End_b Strand_b Distance".split()))


@pytest.fixture()
def hyp1():

    c = """chr1	0	3	+
chr1	1	2	+"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp2():

    c = """chr1	2	3	+"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))

def test_hypothesis_counterexample(hyp1, hyp2, expected_result_counterexample1):

     print(hyp1)
     print(hyp2)

     result = hyp1.nearest(hyp2, how="next", overlap=False)

     print(result)
     # assert 0
     # print(PyRanges(result.df.tail()))
     assert_df_equal(result.df, expected_result_counterexample1.df)



@pytest.fixture()
def expected_result_counterexample2():

    c = """chr1  1  2  +  2   3   +         1"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None,
                                  names="Chromosome Start End Strand Start_b End_b Strand_b Distance".split()))


@pytest.fixture()
def hyp3():

    c = """chr1	0	4	+
chr1	1	2	+"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp4():

    c = """chr1	2	3	+"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))

def test_hypothesis_counterexample2(hyp3, hyp4, expected_result_counterexample2):

     print(hyp3)
     print(hyp4)

     result = hyp3.nearest(hyp4, how="next", overlap=False)

     print(result)
     # assert 0
     # print(PyRanges(result.df.tail()))
     assert_df_equal(result.df, expected_result_counterexample2.df)


@pytest.fixture()
def expected_result_counterexample3():

    c = """chr1 0 1 + 2 3 + 2
chr1 0 2 + 2 3 + 1
chr1 1 2 + 2 3 + 1"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None,
                                  names="Chromosome Start End Strand Start_b End_b Strand_b Distance".split()))



@pytest.fixture()
def hyp5():

    c = """chr1	0	4	+
chr1	0	1	+
chr1	0	2	+
chr1	1	2	+"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp6():

    c = """chr1	2	3	+"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))

def test_hypothesis_counterexample3(hyp5, hyp6, expected_result_counterexample3):

     print(hyp5)
     print(hyp6)

     result = hyp5.nearest(hyp6, how="next", overlap=False)

     print(result)
     # assert 0
     # print(PyRanges(result.df.tail()))
     assert_df_equal(result.df, expected_result_counterexample3.df)



@pytest.fixture()
def hyp7():

    c = """chr1	0	1	+"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp8():

    c = """chr1      0    1      +
chr1      1    2      +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def expected_result_counterexample4():

    c = """chr1  0  1  +  1   2   +         1"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None,
                                  names="Chromosome Start End Strand Start_b End_b Strand_b Distance".split()))

def test_hypothesis_counterexample4(hyp7, hyp8, expected_result_counterexample4):

     print(hyp7)
     print(hyp8)

     result = hyp7.nearest(hyp8, how="next", overlap=False)

     print(result.df)
     # assert 0
     # print(PyRanges(result.df.tail()))

     assert_df_equal(result.df, expected_result_counterexample4.df)



@pytest.fixture()
def minus_chip():

    c = """chr1	9916	10115	HWI-ST216_313:3:1203:10227:6568	1	-
chr1	9951	10150	HWI-ST216_313:3:2205:20086:33508	1	-
chr1	9978	10177	HWI-ST216_313:3:1204:5599:113305	1	-
chr1	10001	10200	HWI-ST216_313:3:1102:14019:151362	1	-
chr1	10127	10326	HWI-ST216_313:3:2207:7406:122346	1	-
chr1	10241	10440	HWI-ST216_313:3:1302:4516:156396	1	-"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Name Score Strand".split()))


@pytest.fixture()
def minus_input():

    c = """chr1	9988	10187	HWI-ST216:427:D29R1ACXX:2:1205:6095:16532	1	-
chr1	10079	10278	HWI-ST216:427:D29R1ACXX:2:1314:10333:38924	1	-
chr1	10082	10281	HWI-ST216:427:D29R1ACXX:2:2213:10285:5151	1	-
chr1	10149	10348	HWI-ST216:427:D29R1ACXX:2:2306:7654:33038	1	-
chr1	19958	20157	HWI-ST216:427:D29R1ACXX:2:1313:6283:67310	1	-"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Name Score Strand".split()))


@pytest.fixture()
def expected_result_minus():

    c = """chr1 9916 10115 HWI-ST216_313:3:1203:10227:6568 1 - 10149 10348 HWI-ST216:427:D29R1ACXX:2:2306:7654:33038 1 - 35
chr1 9951 10150 HWI-ST216_313:3:2205:20086:33508 1 - 19958 20157 HWI-ST216:427:D29R1ACXX:2:1313:6283:67310 1 - 9809
chr1 9978 10177 HWI-ST216_313:3:1204:5599:113305 1 - 19958 20157 HWI-ST216:427:D29R1ACXX:2:1313:6283:67310 1 - 9782
chr1 10001 10200 HWI-ST216_313:3:1102:14019:151362 1 - 19958 20157 HWI-ST216:427:D29R1ACXX:2:1313:6283:67310 1 - 9759
chr1 10127 10326 HWI-ST216_313:3:2207:7406:122346 1 - 19958 20157 HWI-ST216:427:D29R1ACXX:2:1313:6283:67310 1 - 9633
chr1 10241 10440 HWI-ST216_313:3:1302:4516:156396 1 - 19958 20157 HWI-ST216:427:D29R1ACXX:2:1313:6283:67310 1 - 9519"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None,
                                  names="Chromosome Start End Name Score Strand Start_b End_b Name_b Score_b Strand_b Distance".split()))

def test_minus(minus_chip, minus_input, expected_result_minus):

     print(minus_chip)
     print(minus_input)

     result = minus_chip.nearest(minus_input, how="next", overlap=False)

     print(result.df)
     # assert 0
     # print(PyRanges(result.df.tail()))

     assert_df_equal(result.df, expected_result_minus.df)


# @pytest.fixture()
# def plus_input():

#     c = """chr1	10073	10272	HWI-ST216:427:D29R1ACXX:2:2302:14161:85418	1	+
# chr1	10280	10479	HWI-ST216:427:D29R1ACXX:2:2103:15965:69480	1	+
# chr1	16056	16255	HWI-ST216:427:D29R1ACXX:2:1102:7604:12113	1	+
# chr1	16064	16263	HWI-ST216:427:D29R1ACXX:2:2105:18202:77798	1	+
# chr1	16109	16308	HWI-ST216:427:D29R1ACXX:2:2110:12286:25379	1	+"""

#     return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Name Score Strand".split()))



# def test_opposite(minus_chip, plus_input, expected_result_minus):

#      print(minus_chip)
#      print(plus_input)

#      result = minus_chip.nearest(plus_input, how="next", overlap=False)

#      print(result.df)
#      # assert 0
#      # print(PyRanges(result.df.tail()))

#      assert 0
     # assert_df_equal(result.df, expected_result_minus.df)



@pytest.fixture()
def hyp9():

    c = """chr1    2       3       +
chr1    1       4       +
chr1    1       4       +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp10():

    c = """chr1    3       4       +
chr1    4       5       +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def expected_result_counterexample4():

    c = """chr1  1  4  +  4   5   +         1
chr1  1  4  +  4   5   +         1
chr1  2  3  +  3   4   +         1"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None,
                                  names="Chromosome Start End Strand Start_b End_b Strand_b Distance".split()))

def test_hypothesis_counterexample4(hyp9, hyp10, expected_result_counterexample4):

     print(hyp9)
     print(hyp10)

     result = hyp9.nearest(hyp10, how="next", overlap=False)

     print(result.df)
     # assert 0
     # print(PyRanges(result.df.tail()))

     assert_df_equal(result.df, expected_result_counterexample4.df)



@pytest.fixture()
def hyp11():

    c = """chr1    2       3       +
chr1    1       4       +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp12():

    c = """chr1    3       4       +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def expected_result_counterexample5():

    c = """chr1  2  3  +  3   4   +         1"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None,
                                  names="Chromosome Start End Strand Start_b End_b Strand_b Distance".split()))

def test_hypothesis_counterexample5(hyp11, hyp12, expected_result_counterexample5):

     print(hyp11)
     print(hyp12)

     result = hyp11.nearest(hyp12, how="next", overlap=False)

     print(result.df)
     # assert 0
     # print(PyRanges(result.df.tail()))

     assert_df_equal(result.df, expected_result_counterexample5.df)



@pytest.fixture()
def hyp13():

    c = """chr1    1       2       +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp14():

    c = """chr1    1       2       +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))

def test_hypothesis_counterexample6(hyp13, hyp14):

     print(hyp13)
     print(hyp14)

     result = hyp13.nearest(hyp14, how="next", overlap=False)

     print(result.df)
     # assert 0
     # print(PyRanges(result.df.tail()))

     assert result.df.empty



@pytest.fixture()
def expected_result_counterexample6():

    c = """chr1 0 1 + 5726225 5726228 + 5726225
chr1 0 1 + 5726225 5726228 + 5726225
chr1 0 1 + 5726225 5726228 + 5726225
chr1 0 1 + 5726225 5726228 + 5726225
chr1 0 1 + 5726225 5726228 + 5726225
chr1 0 1 + 5726225 5726228 + 5726225
chr1 0 1 + 5726225 5726228 + 5726225
chr1 0 1 + 5726225 5726228 + 5726225
chr1 0 5726225 + 5726225 5726228 + 1
chr1 3538885 3832293 + 5726225 5726228 + 1893933"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None,
                                  names="Chromosome Start End Strand Start_b End_b Strand_b Distance".split()))


@pytest.fixture()
def hyp15():

    c = """chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0  5726225      +
chr1  3538885  3832293      +
chr1  4426346  9655531      +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp16():

    c = """chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1  5726225  5726228      +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))

def test_hypothesis_counterexample6(hyp15, hyp16, expected_result_counterexample6):

     print(hyp15)
     print(hyp16)

     result = hyp15.nearest(hyp16, how="next", overlap=False)

     print(result)
     # assert 0
     # print(PyRanges(result.df.tail()))
     assert_df_equal(result.df, expected_result_counterexample6.df)





import pytest
from tests.helpers import assert_df_equal

from pyranges.pyranges import PyRanges
import pyranges as pr

import pandas as pd

from io import StringIO


@pytest.fixture
def expected_result_unstranded():

    c = """Chromosome Start End Name Score Strand Start_b End_b Name_b Score_b Strand_b Distance
0 chr1 3 6 h 0 + 1 2 f 0 + 2
1 chr1 5 7 h 0 - 1 2 f 0 + 4
2 chr1 8 9 h 0 + 6 7 f 0 - 2"""

    return PyRanges(pd.read_table(StringIO(c), sep=" "))


def test_nearest_bed_unstranded(f1, f2, expected_result_unstranded):

    print(f1)
    print(f2)

    result = f1.nearest(f2, strandedness=False, suffix="_b", how="previous", overlap=False)

    print(result.df.to_csv(sep=" "))

    assert_df_equal(result.df, expected_result_unstranded.df)


@pytest.fixture
def expected_result_opposite_stranded():

    c = """Chromosome Start End Name Score Strand Start_b End_b Name_b Score_b Strand_b Distance
0 chr1 10246 10445 HWI-ST216_313:3:1207:4315:142177 1 + 9988 10187 HWI-ST216:427:D29R1ACXX:2:1205:6095:16532 1 - 60
1 chr1 110246 110445 HWI-ST216_313:3:1207:4315:142177 1 + 19958 20157 HWI-ST216:427:D29R1ACXX:2:1313:6283:67310 1 - 90090"""

    return PyRanges(pd.read_table(StringIO(c), sep=" "))


def test_nearest_bed_opposite_stranded(chip_10_plus_one, input_10, expected_result_opposite_stranded):

    result = chip_10_plus_one.nearest(input_10, strandedness="opposite", suffix="_b", how="previous", overlap=False)

    # print("result", result.df.to_csv(sep=" "))

    assert_df_equal(result.df, expected_result_opposite_stranded.df)


@pytest.fixture
def expected_result_same_stranded():

    c = """Chromosome Start End Name Score Strand Start_b End_b Name_b Score_b Strand_b Distance
0 chr1 10241 10440 HWI-ST216_313:3:1302:4516:156396 1 - 9988 10187 HWI-ST216:427:D29R1ACXX:2:1205:6095:16532 1 - 55
1 chr1 110246 110445 HWI-ST216_313:3:1207:4315:142177 1 + 16109 16308 HWI-ST216:427:D29R1ACXX:2:2110:12286:25379 1 + 93939"""

    return PyRanges(pd.read_table(StringIO(c), sep=" "))


def test_nearest_bed_same_stranded(chip_10_plus_one, input_10, expected_result_same_stranded):

    print(chip_10_plus_one)
    print(input_10)

    result = chip_10_plus_one.nearest(input_10, strandedness="same", suffix="_b", how="previous", overlap=False)

    assert_df_equal(result.df, expected_result_same_stranded.df)




@pytest.fixture()
def expected_result_counterexample1():

    c = """chr1  2  3  +  1   2   +         1"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None,
                                  names="Chromosome Start End Strand Start_b End_b Strand_b Distance".split()))


@pytest.fixture()
def hyp1():

    c = """chr1	2	3	+"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp2():

    c = """chr1	1	2	+
chr1	0	3	+"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))

def test_hypothesis_counterexample(hyp1, hyp2, expected_result_counterexample1):

     print(hyp1)
     print(hyp2)

     result = hyp1.nearest(hyp2, how="previous", overlap=False)

     print(result)
     # 0
     # print(PyRanges(result.df.tail()))
     assert_df_equal(result.df, expected_result_counterexample1.df)



@pytest.fixture()
def expected_result_counterexample2():

    c = """chr1  2  3  +  1   2   +         1"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None,
                                  names="Chromosome Start End Strand Start_b End_b Strand_b Distance".split()))


@pytest.fixture()
def hyp3():

    c = """chr1      2    3      +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp4():

    c = """chr1      1    2      +
chr1      0    4      +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))

def test_hypothesis_counterexample2(hyp3, hyp4, expected_result_counterexample2):

     print(hyp3)
     print(hyp4)

     result = hyp3.nearest(hyp4, how="previous", overlap=False)

     print(result)
     # 0
     # print(PyRanges(result.df.tail()))
     assert_df_equal(result.df, expected_result_counterexample2.df)



@pytest.fixture()
def hyp5():

    c = """chr1      0    1      +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp6():

    c = """chr1      0    1      +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))

def test_hypothesis_counterexample3(hyp5, hyp6):

     print(hyp5)
     print(hyp6)

     result = hyp5.nearest(hyp6, how="previous", overlap=False)

     print(result)
     # 0
     # print(PyRanges(result.df.tail()))
     result.df.empty



@pytest.fixture()
def hyp7():

    c = """chr1	1	2	+
chr1	1	2	+"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp8():

    c = """chr1      0    1      +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def expected_result_counterexample4():

    c = """chr1      1    2      +        0      1        +         1
chr1      1    2      +        0      1        +         1"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None,
                                  names="Chromosome Start End Strand Start_b End_b Strand_b Distance".split()))


def test_hypothesis_counterexample4(hyp7, hyp8, expected_result_counterexample4):

     print(hyp7)
     print(hyp8)

     result = hyp7.nearest(hyp8, how="previous", overlap=False)

     print(result.df)
     # 0
     # print(PyRanges(result.df.tail()))
     assert_df_equal(result.df, expected_result_counterexample4.df)



@pytest.fixture()
def hyp9():

    c = """chr1	0	1	+
chr1	0	1	+"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp10():

    c = """chr1      0    1      +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))

def test_hypothesis_counterexample5(hyp9, hyp10):

     print(hyp9)
     print(hyp10)

     result = hyp9.nearest(hyp10, how="previous", overlap=False)

     print(result.df)
     # 0
     # print(PyRanges(result.df.tail()))
     result.df.empty



@pytest.fixture()
def hyp11():

    c = """chr1	1	2	+
chr1	1	2	+"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp12():

    c = """chr1      0    1      +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def expected_result_counterexample6():

    c = """chr1      1    2      +        0      1        +         1
chr1      1    2      +        0      1        +         1"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None,
                                  names="Chromosome Start End Strand Start_b End_b Strand_b Distance".split()))

def test_hypothesis_counterexample6(hyp11, hyp12, expected_result_counterexample6):

     print(hyp11)
     print(hyp12)

     result = hyp11.nearest(hyp12, how="previous", overlap=False)

     print(result.df)
     # 0
     # print(PyRanges(result.df.tail()))

     assert_df_equal(result.df, expected_result_counterexample6.df)


@pytest.fixture()
def minus_chip():

    c = """chr1	9916	10115	HWI-ST216_313:3:1203:10227:6568	1	-
chr1	9951	10150	HWI-ST216_313:3:2205:20086:33508	1	-
chr1	9978	10177	HWI-ST216_313:3:1204:5599:113305	1	-
chr1	10001	10200	HWI-ST216_313:3:1102:14019:151362	1	-
chr1	10127	10326	HWI-ST216_313:3:2207:7406:122346	1	-
chr1	10241	10440	HWI-ST216_313:3:1302:4516:156396	1	-"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Name Score Strand".split()))


@pytest.fixture()
def minus_input():

    c = """chr1	9988	10187	HWI-ST216:427:D29R1ACXX:2:1205:6095:16532	1	-
chr1	10079	10278	HWI-ST216:427:D29R1ACXX:2:1314:10333:38924	1	-
chr1	10082	10281	HWI-ST216:427:D29R1ACXX:2:2213:10285:5151	1	-
chr1	10149	10348	HWI-ST216:427:D29R1ACXX:2:2306:7654:33038	1	-
chr1	19958	20157	HWI-ST216:427:D29R1ACXX:2:1313:6283:67310	1	-"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Name Score Strand".split()))


@pytest.fixture()
def expected_result_minus():

    c = """0 chr1 10241 10440 HWI-ST216_313:3:1302:4516:156396 1 - 9988 10187 HWI-ST216:427:D29R1ACXX:2:1205:6095:16532 1 - 55"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None,
                                  names="Chromosome Start End Name Score Strand Start_b End_b Name_b Score_b Strand_b Distance".split()))

def test_minus(minus_chip, minus_input, expected_result_minus):

     print(minus_chip)
     print(minus_input)

     result = minus_chip.nearest(minus_input, how="previous", overlap=False)

     print(result.df)
     # 0
     # print(PyRanges(result.df.tail()))

     assert_df_equal(result.df, expected_result_minus.df)



@pytest.fixture()
def hyp13():

    c = """chr1	1	2	+
chr1	1	2	+
chr1	1	2	+
chr1	1	2	+
chr1	1	2	+
chr1	1	2	+
chr1	1	2	+"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp14():

    c = """chr1      0    1      +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def expected_result_counterexample6():

    c = """chr1      1    2      +        0      1        +         1
chr1      1    2      +        0      1        +         1
chr1      1    2      +        0      1        +         1
chr1      1    2      +        0      1        +         1
chr1      1    2      +        0      1        +         1
chr1      1    2      +        0      1        +         1
chr1      1    2      +        0      1        +         1"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None,
                                  names="Chromosome Start End Strand Start_b End_b Strand_b Distance".split()))

def test_hypothesis_counterexample6(hyp13, hyp14, expected_result_counterexample6):

     print(hyp13)
     print(hyp14)

     result = hyp13.nearest(hyp14, how="previous", overlap=False)

     print(result.df)
     # 0
     # print(PyRanges(result.df.tail()))

     assert_df_equal(result.df, expected_result_counterexample6.df)



@pytest.fixture()
def hyp15():

    c = """chr1	0	3	+
chr1	1	2	+"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp16():

    c = """chr1	2	3	+"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


# @pytest.fixture()
# def expected_result_counterexample7():

#     c = """chr1  1  2  +  2   3   +         1"""

#     return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None,
#                                   names="Chromosome Start End Strand Start_b End_b Strand_b Distance".split()))

def test_hypothesis_counterexample7(hyp15, hyp16):

     print(hyp15)
     print(hyp16)

     result = hyp15.nearest(hyp16, how="previous", overlap=False)

     print(result.df)
     # 0
     # print(PyRanges(result.df.tail()))

     result.df.empty # assert_df_equal(result.df, expected_result_counterexample7.df)



@pytest.fixture()
def hyp17():

    c = """chr1      1    2      +
chr1      2    3      +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp18():

    c = """chr1      0    1      +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def expected_result_counterexample8():

    c = """chr1  1  2  +  0   1   +         1
chr1  2  3  +  0   1   +         2"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None,
                                  names="Chromosome Start End Strand Start_b End_b Strand_b Distance".split()))

def test_hypothesis_counterexample8(hyp17, hyp18, expected_result_counterexample8):

     print(hyp17)
     print(hyp18)

     result = hyp17.nearest(hyp18, how="previous", overlap=False)

     print(result.df)
     # 0
     # print(PyRanges(result.df.tail()))

     assert_df_equal(result.df, expected_result_counterexample8.df)



@pytest.fixture()
def hyp19():

    c = """chr1      1    2      +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp20():

    c = """chr1      0    1      +
chr1      0    2      +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def expected_result_counterexample9():

    c = """chr1  1  2  +  0   1   +         1"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None,
                                  names="Chromosome Start End Strand Start_b End_b Strand_b Distance".split()))

def test_hypothesis_counterexample9(hyp19, hyp20, expected_result_counterexample9):

     print(hyp19)
     print(hyp20)

     result = hyp19.nearest(hyp20, how="previous", overlap=False)

     print(result.df)
     # 0
     # print(PyRanges(result.df.tail()))

     assert_df_equal(result.df, expected_result_counterexample9.df)



@pytest.fixture()
def hyp21():

    c = """chr1	1	2	+
chr1	2	3	+"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp22():

    c = """chr1      0    2      +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def expected_result_counterexample10():

    c = """chr1  2  3  +  0   2   +         1"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None,
                                  names="Chromosome Start End Strand Start_b End_b Strand_b Distance".split()))

def test_hypothesis_counterexample10(hyp21, hyp22, expected_result_counterexample10):

     print(hyp21)
     print(hyp22)

     result = hyp21.nearest(hyp22, how="previous", overlap=False)

     print(result.df)
     # 0
     # print(PyRanges(result.df.tail()))

     assert_df_equal(result.df, expected_result_counterexample10.df)



@pytest.fixture()
def hyp23():

    c = """chr1    1       2       +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp24():

    c = """chr1    1       2       +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


def test_hypothesis_counterexample11(hyp23, hyp24):

     print(hyp23)
     print(hyp24)

     result = hyp23.nearest(hyp24, how="previous", overlap=False)

     print(result.df)
     # 0
     # print(PyRanges(result.df.tail()))

     result.df.empty



@pytest.fixture()
def hyp25():

    c = """chr1    3       4       +
chr1    4       5       +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp26():

    c = """chr1    0       4       +
chr1    1       3       +
chr1    2       3       +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def expected_result_counterexample12():

    c = """chr1  3  4  +  2   3   +         1
chr1  4  5  +  0   4   +         1"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None,
                                  names="Chromosome Start End Strand Start_b End_b Strand_b Distance".split()))


def test_hypothesis_counterexample12(hyp25, hyp26, expected_result_counterexample12):

     print(hyp25)
     print(hyp26)

     result = hyp25.nearest(hyp26, how="previous", overlap=False)

     print(result.df)
     # 0
     # print(PyRanges(result.df.tail()))

     assert_df_equal(result.df, expected_result_counterexample12.df)



@pytest.fixture()
def expected_result_counterexample13():

    c = """chr1 3538885 3832293 + 0 1 + 3538885
chr1 4426346 9655531 + 0 1 + 4426346"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None,
                                  names="Chromosome Start End Strand Start_b End_b Strand_b Distance".split()))


@pytest.fixture()
def hyp27():

    c = """chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0  5726225      +
chr1  3538885  3832293      +
chr1  4426346  9655531      +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp28():

    c = """chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1  5726225  5726228      +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))

def test_hypothesis_counterexample13(hyp27, hyp28, expected_result_counterexample13):

     print(hyp27)
     print(hyp2)

     result = hyp27.nearest(hyp28, how="previous", overlap=False)

     print(result)

     assert_df_equal(result.df, expected_result_counterexample13.df)



def get_slice(self, val):

    # 100:999

    start = val.start or 0
    stop = val.stop or max(self.df.End.max(), start)

    idxes = []
    for it in self.__ncls__.values():
        idxes.extend([r[2] for r in it.find_overlap(start, stop)])

    return self.df.loc[idxes]


def get_string(self, val):

    if val in set(self.df.Chromosome):
        return self.df.loc[self.df.Chromosome == val]
    elif val in "+ -".split():
        return self.df.loc[self.df.Strand == val]
    else:
        raise Exception("Chromosome or Strand '{}' not found in PyRanges.".format(val))



def get_tuple(self, val):

    if len(val) == 2:
        df = get_double(self, val)
    elif len(val) == 3:
        df = get_triple(self, val)

    return df


def get_double(self, val):

    # "chr1", 5:10
    if len(val) == 2 and val[0] in self.df.Chromosome.values and isinstance(val[1], slice):
        chromosome, loc = val
        start = loc.start or 0
        stop = loc.stop or max(self.df.loc[self.df.Chromosome == chromosome].End.max(), start)
        idxes = [r[2] for r in self.__ncls__[chromosome, "+"].find_overlap(start, stop)] + \
                [r[2] for r in self.__ncls__[chromosome, "-"].find_overlap(start, stop)]

        return self.df.loc[idxes]

    # "+", 5:10
    if len(val) == 2 and val[0] in "+ -".split() and isinstance(val[1], slice):
        strand, loc = val
        start = loc.start or 0
        stop = loc.stop or max(self.df.loc[self.df.Chromosome == chromosome].End.max(), start)
        idxes = []
        for chromosome in self.df.Chromosome.drop_duplicates():
            idxes.extend([r[2] for r in self.__ncls__[chromosome, strand].find_overlap(start, stop)])

        return self.df.loc[idxes]

    # "chr1", "+"
    if len(val) == 2 and val[1] in "+ -".split():

        chromosome, strand = val

        return self.df.loc[(self.df.Chromosome == chromosome) & (self.df.Strand == strand)]


def get_triple(self, val):

    # "chr1", "+", 5:10
    chromosome, strand, loc = val
    start = loc.start or 0
    stop = loc.stop or max(self.df.loc[(self.df.Chromosome == chromosome) & (self.df.Strand == strand)].End.max(), start)
    if strand not in "+ -".split():
        raise Exception("Strand '{}' invalid.".format(val))

    if (chromosome, strand) in self.__ncls__:
        idxes = [r[2] for r in self.__ncls__[chromosome, strand].find_overlap(start, stop)]
        return self.df.loc[idxes]
    else:
        raise Exception("Chromosome and Strand pair {}, {} not found in PyRanges.".format(chromosome, strand))



import pandas as pd
import numpy as np
from collections import defaultdict

from tabulate import tabulate


from pyranges.genomicfeatures import GenomicFeaturesMethods
# from pyranges.genomicfeatures import _tss, _tes
from pyranges.subset import get_string, get_slice, get_tuple
from pyranges.methods import _cluster, _subtraction, _set_union, _set_intersection, _intersection, _nearest, _coverage, _overlap_write_both, _overlap, _tss, _tes, _jaccard, _lengths, _slack
# from pyranges.multithreaded import _cluster, pyrange_apply_single, _subtraction, _set_union, _set_intersection, _intersection, pyrange_apply, _nearest, _coverage, _write_both, _first_df

from ncls import NCLS

def return_copy_if_view(df, df2):
    # https://stackoverflow.com/questions/26879073/checking-whether-data-frame-is-copy-or-view-in-pandas

    if df.values.base is df2.values.base:
        return df.copy(deep=True)
    else:
        return df

# def is_view(df, df2):
#     # https://stackoverflow.com/questions/26879073/checking-whether-data-frame-is-copy-or-view-in-pandas

#     return df.values.base is df2.values.base

def create_ncls(cdf):

    return NCLS(cdf.Start.values,
                cdf.End.values,
                cdf.index.values)


def create_ncls_dict(df, strand):

    if not strand:
        grpby_key = "Chromosome"
    else:
        grpby_key = "Chromosome Strand".split()

    grpby = df.groupby(grpby_key)
    nclses = {key: create_ncls(cdf) for (key, cdf) in grpby}
    dd = defaultdict(NCLS)
    dd.update(nclses)

    return dd


def create_pyranges_df(seqnames, starts, ends, strands=None):

    if isinstance(seqnames, str):
        seqnames = pd.Series([seqnames] * len(starts), dtype="category")

    if strands != None and strands != False:

        if isinstance(strands, str):
            strands = pd.Series([strands] * len(starts), dtype="category")

        columns = [seqnames, starts, ends, strands]
        lengths = list(str(len(s)) for s in columns)
        assert len(set(lengths)) == 1, "seqnames, starts, ends and strands must be of equal length. But are {}".format(", ".join(lengths))
        colnames = "Chromosome Start End Strand".split()
    else:
        columns = [seqnames, starts, ends]
        lengths = list(str(len(s)) for s in columns)
        assert len(set(lengths)) == 1, "seqnames, starts and ends must be of equal length. But are {}".format(", ".join(lengths))
        colnames = "Chromosome Start End".split()

    idx = range(len(starts))
    series_to_concat = []
    for s in columns:
        s = pd.Series(s, index=idx)
        series_to_concat.append(s)

    df = pd.concat(series_to_concat, axis=1)
    df.columns = colnames

    return df


def return_empty_if_one_empty(func):

    def extended_func(self, other, *args, **kwargs):

        if len(self) == 0 or len(other) == 0:
            df = pd.DataFrame(columns="Chromosome Start End Strand".split())
        else:
            df = func(self, other, *args, **kwargs)

        return df

    return extended_func


def return_empty_if_both_empty(func):

    def extended_func(self, other, *args, **kwargs):

        if len(self) == 0 and len(other) == 0:
            df = pd.DataFrame(columns="Chromosome Start End Strand".split())
        else:
            df = func(self, other, *args, **kwargs)

        return df

    return extended_func


def pyrange_or_df(func):

    def extension(self, other, *args, **kwargs):
        df = func(self, other, *args, **kwargs)

        if kwargs.get("df"):
            return df

        return PyRanges(df)

    return extension


def pyrange_or_df_single(func):

    def extension(self, *args, **kwargs):
        df = func(self, *args, **kwargs)

        if kwargs.get("df"):
            return df

        return PyRanges(df)

    return extension



class PyRanges():

    df = None
    gf = None

    def __init__(self, df=None, seqnames=None, starts=None, ends=None, strands=None, copy_df=True):

        df_given = True if not df is None else False

        if df is False or df is None:
            df = create_pyranges_df(seqnames, starts, ends, strands)
        else:
            assert "Chromosome" in df and "Start" in df and "End" in df
            df.index = range(len(df))

        # using __dict__ to avoid invoking setattr
        if "Strand" not in df:
            self.__dict__["stranded"] = False
        else:
            self.__dict__["stranded"] = True

        if copy_df and df_given:
            df = df.copy()
            df.Chromosome = df.Chromosome.astype("category")
            if "Strand" in df:
                df.Strand = df.Strand.astype("category")


        self.__dict__["df"] = df

        self.__dict__["__ncls__"] = create_ncls_dict(df, self.stranded)

        self.__dict__["ft"] = GenomicFeaturesMethods(self)


    def __len__(self):
        return self.df.shape[0]

    def __setattr__(self, column_name, column):

        if column_name in "Chromosome Start End Strand".split():
            raise Exception("The columns Chromosome, Start, End or Strand can not be reset.")
        if column_name == "stranded":
            raise Exception("The stranded attribute is read-only. Create a new PyRanges object instead.")

        if column_name == "df":
            self.__dict__["df"] = column
            return

        if not isinstance(column, str):
            if not len(self.df) == len(column):
                raise Exception("DataFrame and column must be same length.")

            column_to_insert = pd.Series(column, index=self.df.index)
        else:
            column_to_insert = pd.Series(column, index=self.df.index)

        pos = self.df.shape[1]
        if column_name in self.df:
            pos = list(self.df.columns).index(column_name)
            self.df.drop(column_name, inplace=True, axis=1)

        self.df.insert(pos, column_name, column_to_insert)


    def __getattr__(self, name):

        if name in self.df:
            return self.df[name]
        else:
            self.__dict__[name]

    def __eq__(self, other):

        return self.df.equals(other.df)

    def __getitem__(self, val):

        if isinstance(val, str):
            df = get_string(self, val)

        elif isinstance(val, tuple):
            df = get_tuple(self, val)
        elif isinstance(val, slice):
            df = get_slice(self, val)

        if not df._is_view:
            return PyRanges(df)
        else:
            return PyRanges(df.copy(deep=True))


    def __str__(self):

        if len(self.df) > 6:
            h = self.df.head(3).astype(object)
            t = self.df.tail(3).astype(object)
            m = self.df.head(1).astype(object)
            m.loc[:,:] = "..."
            m.index = ["..."]
            s = pd.concat([h, m, t])
        else:
            s = self.df

        h = [c + "\n(" + str(t) + ")" for c, t in  zip(self.df.columns, self.df.dtypes)]

        str_repr = tabulate(s, headers=h, tablefmt='psql', showindex=False) + \
                                        "\nPyRanges object has {} sequences from {} chromosomes.".format(self.df.shape[0], len(set(self.df.Chromosome)))
        return str_repr


    def __repr__(self):

        return str(self)


    @pyrange_or_df
    @return_empty_if_one_empty
    def overlap(self, other, strandedness=False, invert=False, how=None, **kwargs):

        "Want all intervals in self that overlap with other."

        df = _overlap(self, other, strandedness, invert, how)

        return df

    @pyrange_or_df
    @return_empty_if_one_empty
    def nearest(self, other, strandedness=False, suffix="_b", how=None, overlap=True, **kwargs):

        "Find the nearest feature in other."

        df = _nearest(self, other, strandedness, suffix, how, overlap)

        return df

    @pyrange_or_df
    @return_empty_if_one_empty
    def intersection(self, other, strandedness=False, how=None):

        "Want the parts of the intervals in self that overlap with other."

        df = _intersection(self, other, strandedness, how)

        return df

    @pyrange_or_df
    @return_empty_if_one_empty
    def set_intersection(self, other, strandedness=False, how=None):

        si = _set_intersection(self, other, strandedness, how)

        return si

    @pyrange_or_df
    @return_empty_if_both_empty
    def set_union(self, other, strand=False):

        si = _set_union(self, other, strand)

        return si


    @pyrange_or_df
    def subtraction(self, other, strandedness=False):

        return _subtraction(self, other, strandedness)


    @pyrange_or_df
    @return_empty_if_one_empty
    def join(self, other, strandedness=False, new_pos=None, suffixes=["_a", "_b"], suffix="_b", how=None):

        df = _overlap_write_both(self, other, strandedness, new_pos, suffixes, suffix, how)

        return df


    @pyrange_or_df_single
    def cluster(self, strand=None, max_dist=0, min_nb=1, **kwargs):

        df = _cluster(self, strand, max_dist, min_nb)

        return df

    # @pyrange_or_df_single
    # def tile(self, tile_size=50):

    #     df = _tile(self, tile_size)
    #     return df


    def coverage(self, value_col=None, stranded=False):

        return _coverage(self, value_col, stranded=stranded)


    @pyrange_or_df_single
    def slack(self, slack):

        return _slack(self, slack)



    @pyrange_or_df_single
    def tssify(self, slack=0):

        if not self.stranded:
            raise Exception("Cannot compute TSSes without strand info. Perhaps use slack() instead?")

        return _tss(self, slack)


    @pyrange_or_df_single
    def tesify(self, slack=0):

        return _tes(self, slack)

    def pos(self, *val):

        # turn int-tuple into slice
        newval = []
        for v in val:
            if isinstance(v, tuple) and len(v) == 2 and isinstance(v[0], int) and isinstance(v[1], int):
                newval.append(slice(v[0], v[1]))
            else:
                newval.append(v)

        val = tuple(newval)

        if len(val) == 1:
            val = val[0]

        if isinstance(val, str):
            df = get_string(self, val)
        elif isinstance(val, tuple):
            df = get_tuple(self, val)
        elif isinstance(val, slice):
            df = get_slice(self, val)
        else:
            raise Exception("Invalid type for indexer. Must be str, slice, or 2/3-tuple.")

        return return_copy_if_view(df, self.df)


    # always returns df
    def get(self, column, values, **kwargs):

        df = self.df
        if not column in self.df:
            raise Exception("Column {} not in PyRanges".format(column))

        if isinstance(values, str):
            df = df.loc[df[column] == values]
        else:
            df = df.loc[df[column].isin(values)]

        return return_copy_if_view(df, self.df)


    def lengths(self, **kwargs):

        df = _lengths(self)

        return df


    def jaccard(self, other, strandedness):

        return _jaccard(self, other, strandedness)


    @pyrange_or_df_single
    def sort(self, strand=True):

        if strand:
            return self.df.sort_values("Chromosome Strand".split())
        else:
            return self.df.sort_values("Chromosome")



    # @pyrange_or_df
    # @return_empty_if_one_empty
    # def overlap(self, other, strandedness=False, invert=False, how=None, nb_cpu=1, **kwargs):

    #     "Want all intervals in self that overlap with other."

    #     df = pyrange_apply(_first_df, self, other, strandedness=strandedness,
    #                        invert=invert, how=how, n_jobs=nb_cpu, **kwargs)

    #     return df

    # @pyrange_or_df
    # @return_empty_if_one_empty
    # def nearest(self, other, strandedness=False, suffix="_b", how=None, overlap=True, nb_cpu=1, **kwargs):
    #     "Find the nearest feature in other."

    #     df = pyrange_apply(_nearest, self, other, strandedness=strandedness, suffix=suffix, how=how, overlap=overlap, n_jobs=nb_cpu)

    #     return df

    # @pyrange_or_df
    # @return_empty_if_one_empty
    # def intersection(self, other, strandedness=False, how=None, nb_cpu=1):

    #     "Want the parts of the intervals in self that overlap with other."

    #     df = pyrange_apply(_intersection, self, other, strandedness=strandedness, how=how, n_jobs=nb_cpu)
    #     # df = _intersection(self, other, strandedness=strandedness, how=how)

    #     return df

    # @pyrange_or_df
    # @return_empty_if_one_empty
    # def set_intersection(self, other, strandedness=False, how=None, nb_cpu=1):

    #     si = pyrange_apply(_set_intersection, self, other, strandedness=strandedness, how=how, n_jobs=nb_cpu)

    #     return si

    # @pyrange_or_df
    # @return_empty_if_both_empty
    # def set_union(self, other, strandedness=False, nb_cpu=1):

    #     si = pyrange_apply(_set_union, self, other, strandedness=strandedness, n_jobs=nb_cpu)

    #     return si


    # @pyrange_or_df
    # def subtraction(self, other, strandedness=False, nb_cpu=1):


    #     return pyrange_apply(_subtraction, self, other, strandedness=strandedness, n_jobs=nb_cpu)


    # @pyrange_or_df
    # @return_empty_if_one_empty
    # def join(self, other, strandedness=False, new_pos=None, suffixes=["_a", "_b"], how=None, nb_cpu=1, **kwargs):

    #     df = pyrange_apply(_write_both, self, other, strandedness=strandedness, new_pos=new_pos,
    #                        suffixes=suffixes, how=how, n_jobs=nb_cpu, **kwargs)

    #     return df


    # @pyrange_or_df_single
    # def cluster(self, strand=None, nb_cpu=1):

    #     df = pyrange_apply_single(_cluster, self, strand=strand, n_jobs=nb_cpu)

    #     return df


    # @pyrange_or_df_single
    # def tile(self, tile_size=50, nb_cpu=1):

    #     df = _tile(self, tile_size)
    #     return df


    # def coverage(self, value_col=None, stranded=False, nb_cpu=1):

    #     return _coverage(self, value_col, stranded=stranded, n_jobs=nb_cpu)





import pytest
from tests.helpers import assert_df_equal

from pyranges.pyranges import PyRanges
import pyranges as pr

import pandas as pd

from io import StringIO


@pytest.fixture()
def expected_result_counterexample1():

    c = """chr1	0	1	+	0	1	+	0
chr1	0	1	+	0	1	+	0
chr1	0	1	+	0	1	+	0
chr1	0	1	+	0	1	+	0
chr1	0	1	+	0	1	+	0
chr1	0	1	+	0	1	+	0
chr1	0	1	+	0	1	+	0
chr1	0	1	+	0	1	+	0
chr1	0	5726225	+	0	1	+	0
chr1	3538885	3832293	+	5726225	5726228	+	1893933
chr1	4426346	9655531	+	5726225	5726228	+	0"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None,
                                  names="Chromosome Start End Strand Start_b End_b Strand_b Distance".split()))


@pytest.fixture()
def hyp1():

    c = """chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0  5726225      +
chr1  3538885  3832293      +
chr1  4426346  9655531      +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))


@pytest.fixture()
def hyp2():

    c = """chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1        0        1      +
chr1  5726225  5726228      +"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None, names="Chromosome Start End Strand".split()))

def test_hypothesis_counterexample(hyp1, hyp2, expected_result_counterexample1):

     print(hyp1)
     print(hyp2)

     result = hyp1.nearest(hyp2, overlap=True)

     print(result)
     # assert 0
     # print(PyRanges(result.df.tail()))
     assert_df_equal(result.df, expected_result_counterexample1.df)


def test_hypothesis_counterexample_nearest_next(hyp1, hyp2, expected_result_counterexample1):

     print(hyp1)
     print(hyp2)

     result = hyp1.nearest(hyp2, overlap=True, how="next")

     print(result)
     # print(PyRanges(result.df.tail()))
     assert_df_equal(result.df, expected_result_counterexample1.df)


@pytest.fixture()
def expected_result_nearest_previous():

    c = """chr1 0 1 + 0 1 + 0
chr1 0 1 + 0 1 + 0
chr1 0 1 + 0 1 + 0
chr1 0 1 + 0 1 + 0
chr1 0 1 + 0 1 + 0
chr1 0 1 + 0 1 + 0
chr1 0 1 + 0 1 + 0
chr1 0 1 + 0 1 + 0
chr1 0 5726225 + 0 1 + 0
chr1 3538885 3832293 + 0 1 + 3538885
chr1 4426346 9655531 + 5726225 5726228 + 0"""

    return PyRanges(pd.read_table(StringIO(c), sep="\s+", header=None,
                                  names="Chromosome Start End Strand Start_b End_b Strand_b Distance".split()))


def test_hypothesis_counterexample_nearest_previous(hyp1, hyp2, expected_result_nearest_previous):

     print(hyp1)
     print(hyp2)

     result = hyp1.nearest(hyp2, overlap=True, how="previous")

     print(result)
     # print(PyRanges(result.df.tail()))
     assert_df_equal(result.df, expected_result_nearest_previous.df)



import numpy as np
import pandas as pd
from ncls import NCLS

import pyranges as pr

from natsort import natsorted

from joblib import Parallel, delayed

from sorted_nearest import (find_clusters, nearest_previous_nonoverlapping,
                            nearest_next_nonoverlapping, nearest_nonoverlapping, find_clusters)

from collections import defaultdict

from functools import wraps

def parse_grpby_key(grpby_key):

    if isinstance(grpby_key, str):
        return grpby_key, False
    else:
        return grpby_key[0], grpby_key[1]


def return_empty_if_one_empty(func):

    @wraps(func)
    def extended_func(self, other, **kwargs):

        if len(self) == 0 or len(other) == 0:
            df = pd.DataFrame(columns="Chromosome Start End".split())
        else:
            df = func(self, other, **kwargs)

        return df

    return extended_func


def _pyrange_apply(function, scdf, other_dfs, grpby_key, n_jobs=1, **kwargs):

    print("in pyrange apply " * 10, n_jobs)
    if function.__name__ == "_set_union":
        self_dfs =  {k: d for k, d in scdf.groupby(grpby_key)}
        self_dfs = defaultdict(lambda: pd.DataFrame(columns="Chromosome Start End".split()), self_dfs)
        keys_union = natsorted(list(set(self_dfs).union(other_dfs)))
        outdfs = Parallel(n_jobs=n_jobs)(delayed(function)(self_dfs[key], other_dfs[key], key=key, n_jobs=n_jobs, **kwargs) for key in keys_union)

    elif not function.__name__ in ["_write_both", "_nearest"]:
        # for most methods, we do not need to know anything about other except start, end
        # i.e. less data that needs to be sent to other processes
        outdfs = Parallel(n_jobs=n_jobs)(delayed(function)(scdf, other_dfs[key][["Start", "End"]], key=key, n_jobs=n_jobs, **kwargs) for key, scdf in natsorted(scdf.groupby(grpby_key)))

    else:
        outdfs = Parallel(n_jobs=n_jobs)(delayed(function)(scdf, other_dfs[key], key=key, n_jobs=n_jobs, **kwargs) for key, scdf in natsorted(scdf.groupby(grpby_key)))

    outdfs = [df for df in outdfs if not df.empty]

    if outdfs:
        df = pd.concat(outdfs)
        return df
    else:
        return pd.DataFrame(columns="Chromosome Start End Strand".split())


def pyrange_apply_single(function, self, **kwargs):

    strand = kwargs["strand"]
    n_jobs = kwargs.get("n_jobs", 1)

    print("Using {} cores and strand {}".format(n_jobs, strand))

    if strand:
        assert self.stranded, \
            "Can only do stranded operation when PyRange contains strand info"

    if self.stranded and strand:
        grpby_key = ["Chromosome", "Strand"]
    else:
        grpby_key = "Chromosome"

    outdfs = Parallel(n_jobs=n_jobs)(delayed(function)(scdf, **kwargs) for key, scdf in natsorted(self.df.groupby(grpby_key)))


    outdfs = [df for df in outdfs if not df.empty]

    if outdfs:
        df = pd.concat(outdfs)
        return df
    else:
        return pd.DataFrame(columns="Chromosome Start End Strand".split())


def pyrange_apply(function, self, other, n_jobs, **kwargs):

    strandedness = kwargs["strandedness"]

    print("Using {} cores and strandedness {}".format(n_jobs, strandedness))

    assert strandedness in ["same", "opposite", False, None]

    if strandedness:
        assert self.stranded and other.stranded, \
            "Can only do stranded operations when both PyRanges contain strand info"

    if self.stranded and other.stranded and strandedness:
        grpby_key = ["Chromosome", "Strand"]
    else:
        grpby_key = "Chromosome"

    other_strand = {"+": "-", "-": "+"}
    if strandedness == "opposite":
        other_dfs = {(c, other_strand[s]): v for (c, s), v in other.df.groupby(grpby_key)}
        other_dfs = defaultdict(lambda: pd.DataFrame(columns="Chromosome Start End Strand".split()), other_dfs)
    else:
        other_dfs = {key: v for key, v in other.df.groupby(grpby_key)}
        other_dfs = defaultdict(lambda: pd.DataFrame(columns="Chromosome Start End".split()), other_dfs)

    return _pyrange_apply(function, self.df, other_dfs, grpby_key, n_jobs=n_jobs, **kwargs)




@return_empty_if_one_empty
def _first_df(scdf, ocdf, how=False, invert=False, n_jobs=1, **kwargs):

    assert how in "containment first".split() + [False, None]
    starts = scdf.Start.values
    ends = scdf.End.values
    indexes = scdf.index.values

    print("n_jobs " * 10)
    print(n_jobs)

    if n_jobs > 1:
        print("deepcopy")
        scdf = scdf.copy(deep=True)

    it = NCLS(ocdf.Start.values, ocdf.End.values, ocdf.index.values)

    if not how:
        _indexes = it.has_overlaps(starts, ends, indexes)
    elif how == "containment":
        _indexes = it.has_containments(starts, ends, indexes)

    if not invert:
        return scdf.reindex(_indexes)
    else:
        return scdf.loc[~scdf.index.isin(_indexes)]


def _both_dfs(scdf, ocdf, how=False, **kwargs):

    assert how in "containment first".split() + [False, None]
    starts = scdf.Start.values
    ends = scdf.End.values
    indexes = scdf.index.values

    it = NCLS(ocdf.Start.values, ocdf.End.values, ocdf.index.values)

    if not how:
        _self_indexes, _other_indexes = it.all_overlaps_both(starts, ends, indexes)
    elif how == "containment":
        _self_indexes, _other_indexes = it.all_containments_both(starts, ends, indexes)
    else:
        _self_indexes, _other_indexes = it.first_overlap_both(starts, ends, indexes)

    _self_indexes = _self_indexes
    _other_indexes = _other_indexes

    return scdf.loc[_self_indexes], ocdf.loc[_other_indexes]




@return_empty_if_one_empty
def _intersection(scdf, ocdf, **kwargs):

    scdf, ocdf = _both_dfs(scdf, ocdf, **kwargs)

    new_starts = pd.Series(
        np.where(scdf.Start.values > ocdf.Start.values, scdf.Start, ocdf.Start),
        index=scdf.index, dtype=np.long)

    new_ends = pd.Series(
        np.where(scdf.End.values < ocdf.End.values, scdf.End, ocdf.End),
        index=scdf.index, dtype=np.long)

    pd.options.mode.chained_assignment = None  # default='warn'
    scdf.loc[:, "Start"] = new_starts
    scdf.loc[:, "End"] = new_ends
    pd.options.mode.chained_assignment = 'warn'

    return scdf


def _create_df_from_starts_ends(starts, ends, chromosome, strand=None):

    nidx = pd.Index(range(len(starts)))
    if strand:
        cluster_df = pd.DataFrame({"Chromosome": pd.Series(chromosome, dtype="category", index=nidx),
                                   "Start": starts, "End": ends,
                                   "Strand": pd.Series(strand, dtype="category", index=nidx)})
    else:
        cluster_df = pd.DataFrame({"Chromosome": pd.Series(chromosome, dtype="category", index=nidx), "Start": starts, "End": ends})

    return cluster_df


def _cluster(df, chromosome, strand=False, **kwargs):

    cdf = df.sort_values("Start")
    starts, ends = find_clusters(cdf.Start.values, cdf.End.values)

    cluster_df = _create_df_from_starts_ends(starts, ends, chromosome, strand)

    return cluster_df


@return_empty_if_one_empty
def _set_intersection(scdf, ocdf, strandedness=None, how=None, **kwargs):

    chromosome, strand = parse_grpby_key(kwargs["key"])

    strand = True if strandedness else False
    s = _cluster(scdf, chromosome, strand=strand)
    o = _cluster(ocdf, chromosome, strand=strand)

    return _intersection(s, o, strandedness=strandedness, how=how, **kwargs)


def _overlapping_for_nearest(scdf, ocdf, suffix, n_jobs=1, **kwargs):

    nearest_df = pd.DataFrame(columns="Chromosome Start End Strand".split())

    scdf2, ocdf2 = _both_dfs(scdf, ocdf, how="first")

    if not ocdf2.empty:
        # only copying data because of the eternal source buffer array is read only problem
        if n_jobs > 1:
            scdf = scdf.copy(deep=True)
            original_idx = scdf.index.copy(deep=True)
        else:
            original_idx = scdf.index

        idxs = scdf2.index
        original_idx = scdf.index.copy(deep=True)
        missing_idxs = ~original_idx.isin(idxs)
        missing_overlap = scdf.index[missing_idxs]

        df_to_find_nearest_in = scdf.reindex(missing_overlap)

        odf = ocdf.reindex(ocdf2.index)
        odf.index = idxs
        sdf = scdf.reindex(idxs)

        nearest_df = sdf.join(odf, rsuffix=suffix)
        nearest_df.insert(nearest_df.shape[1], "Distance", 0)
    else:
        df_to_find_nearest_in = scdf

    return nearest_df, df_to_find_nearest_in


def _next_nonoverlapping(left_ends, right_starts, right_indexes):

    left_ends = left_ends.sort_values()
    right_starts = right_starts.sort_values()
    r_idx, dist = nearest_next_nonoverlapping(left_ends.values - 1, right_starts.values, right_indexes)
    r_idx = pd.Series(r_idx, index=left_ends.index).sort_index().values
    dist = pd.Series(dist, index=left_ends.index).sort_index().values

    return r_idx, dist


def _previous_nonoverlapping(left_starts, right_ends):

    left_starts = left_starts.sort_values()
    right_ends = right_ends.sort_values()
    r_idx, dist = nearest_previous_nonoverlapping(left_starts.values, right_ends.values - 1, right_ends.index.values)

    r_idx = pd.Series(r_idx, index=left_starts.index).sort_index().values
    dist = pd.Series(dist, index=left_starts.index).sort_index().values

    return r_idx, dist

def sort_one_by_one(d, col1, col2):
    """
    Equivalent to pd.sort_values(by=[col1, col2]), but faster.
    """
    d = d.sort_values(by=[col2])
    return d.sort_values(by=[col1], kind='mergesort')


@return_empty_if_one_empty
def _nearest(scdf, ocdf, suffix="_b", how=None, overlap=True, **kwargs):

    if overlap:
        nearest_df, df_to_find_nearest_in = _overlapping_for_nearest(scdf, ocdf, suffix, **kwargs)
    else:
        df_to_find_nearest_in = scdf

    df_to_find_nearest_in = sort_one_by_one(df_to_find_nearest_in, "Start", "End")
    ocdf = sort_one_by_one(ocdf, "Start", "End")
    df_to_find_nearest_in.index = pd.Index(range(len(df_to_find_nearest_in)))

    if how == "next":
        r_idx, dist = _next_nonoverlapping(df_to_find_nearest_in.End, ocdf.Start, ocdf.index.values)
    elif how == "previous":
        r_idx, dist = _previous_nonoverlapping(df_to_find_nearest_in.Start, ocdf.End)
    else:
        previous_r_idx, previous_dist = _previous_nonoverlapping(df_to_find_nearest_in.Start, ocdf.End)

        next_r_idx, next_dist = _next_nonoverlapping(df_to_find_nearest_in.End, ocdf.Start, ocdf.index.values)

        r_idx, dist = nearest_nonoverlapping(previous_r_idx,
                                             previous_dist,
                                             next_r_idx, next_dist)

    ocdf = ocdf.reindex(r_idx, fill_value=-1) # instead of np.nan, so ints are not promoted to float

    ocdf.index = df_to_find_nearest_in.index
    ocdf.insert(ocdf.shape[1], "Distance", pd.Series(dist, index=ocdf.index).fillna(-1).astype(int))

    r_idx = pd.Series(r_idx, index=ocdf.index)
    df_to_find_nearest_in = df_to_find_nearest_in.drop(r_idx.loc[r_idx == -1].index)

    df = df_to_find_nearest_in.join(ocdf, rsuffix=suffix)

    if overlap and not df.empty and not nearest_df.empty:
        df = pd.concat([nearest_df, df])
    elif overlap and not nearest_df.empty:
        df = nearest_df

    df = df.drop("Chromosome" + suffix, axis=1)
    return df


def _set_union(scdf, ocdf, **kwargs):

    chromosome, strand = parse_grpby_key(kwargs["key"])

    strandedness = kwargs["strandedness"]
    strand = True if strandedness == "same" else False

    if len(scdf) == 0:
        return _cluster(ocdf, chromosome, strand=strand)
    elif len(ocdf) == 0:
        return _cluster(scdf, chromosome, strand=strand)

    _starts = np.concatenate([
        scdf.Start.values,
        ocdf.Start.values])
    _ends = np.concatenate([
        scdf.End.values,
        ocdf.End.values])

    cdf = pd.DataFrame({"Start": _starts, "End": _ends})["Start End".split()]
    cdf = cdf.sort_values("Start")
    starts, ends = find_clusters(cdf.Start.values, cdf.End.values)

    chromosome = scdf.head(1)["Chromosome"].iloc[0]
    if strandedness == "same":
        _strand = scdf.head(1)["Strand"].iloc[0]
    else:
        _strand = False

    cluster_df = _create_df_from_starts_ends(starts, ends, chromosome, _strand)

    return cluster_df


def _subtraction(scdf, ocdf, **kwargs):

    chromosome, strand = parse_grpby_key(kwargs["key"])

    if ocdf.empty or scdf.empty:
        return scdf

    strandedness = kwargs["strandedness"]
    strand = True if strandedness else False

    oc = _cluster(ocdf, chromosome, strand)
    o = NCLS(oc.Start.values, oc.End.values, oc.index.values)

    idx_self, new_starts, new_ends = o.set_difference_helper(
        scdf.Start.values,
        scdf.End.values,
        scdf.index.values)

    missing_idx = pd.Index(scdf.index).difference(idx_self)

    idx_to_drop = new_starts != -1

    new_starts = new_starts[idx_to_drop]
    new_ends = new_ends[idx_to_drop]

    idx_self = idx_self[idx_to_drop]
    new_starts = pd.Series(new_starts, index=idx_self).sort_index()
    new_ends = pd.Series(new_ends, index=idx_self).sort_index()
    idx_self = np.sort(idx_self)

    scdf = scdf.reindex(missing_idx.union(idx_self))

    if len(idx_self):

        scdf.loc[scdf.index.isin(idx_self), "Start"] = new_starts
        scdf.loc[scdf.index.isin(idx_self), "End"] = new_ends

    return scdf


def _coverage(ranges, value_col=None, stranded=False, n_jobs=1, **coverage):

    try:
        from pyranges import PyRles
    except ImportError:
        raise Exception("Using the coverage method requires that pyrle is installed.")

    return PyRles(ranges, value_col=value_col, stranded=stranded, nb_cpu=n_jobs)



@return_empty_if_one_empty
def _write_both(scdf, ocdf, new_pos=False, **kwargs):

    suffixes = kwargs["suffixes"]

    scdf, ocdf = _both_dfs(scdf, ocdf, **kwargs)
    nix = pd.Index(range(len(scdf)))
    scdf.index = nix
    ocdf.index = nix

    ocdf = ocdf.drop("Chromosome", axis=1)

    if not new_pos:
        df = scdf.join(ocdf, rsuffix=suffixes[1])

    elif new_pos == "intersection":

        new_starts = pd.Series(
            np.where(scdf.Start.values > ocdf.Start.values, scdf.Start, ocdf.Start),
            index=scdf.index, dtype=np.long)

        new_ends = pd.Series(
            np.where(scdf.End.values < ocdf.End.values, scdf.End, ocdf.End),
            index=scdf.index, dtype=np.long)
        df = scdf.join(ocdf, lsuffix=suffixes[0], rsuffix=suffixes[1])
        df.insert(1, "Start", new_starts)
        df.insert(2, "End", new_ends)
        df.rename(index=str, columns={"Chromosome" + suffixes[0]: "Chromosome", "Strand" + suffixes[0]: "Strand"}, inplace=True)

    elif new_pos == "union":

        new_starts = pd.Series(
            np.where(scdf.Start.values < ocdf.Start.values, scdf.Start, ocdf.Start),
            index=scdf.index, dtype=np.long)

        new_ends = pd.Series(
            np.where(scdf.End.values > ocdf.End.values, scdf.End, ocdf.End),
            index=scdf.index, dtype=np.long)
        df = scdf.join(ocdf, lsuffix=suffixes[0], rsuffix=suffixes[1])
        df.insert(1, "Start", new_starts)
        df.insert(2, "End", new_ends)
        df.rename(index=str, columns={"Chromosome" + suffixes[0]: "Chromosome", "Strand" + suffixes[0]: "Strand"}, inplace=True)

    return df


if __name__ == "__main__":

    chip_f = "/mnt/scratch/endrebak/pyranges_benchmark/data/download/chip_1000000.bed.gz"
    background_f = "/mnt/scratch/endrebak/pyranges_benchmark/data/download/input_1000000.bed.gz"


    chip = pd.read_table(chip_f, sep="\t", usecols=[0, 1, 2, 5], header=None,
                            names="Chromosome Start End Strand".split(),
                            dtype={"Chromosome": "category", "Strand": "category"} )

    cgr = pr.PyRanges(chip, copy_df=False)

    background = pd.read_table(background_f, sep="\t", usecols=[0, 1, 2, 5],
                                header=None, names="Chromosome Start End Strand".split(),
                                dtype={"Chromosome": "category", "Strand": "category"})

    bgr = pr.PyRanges(background, copy_df=False)



import pandas as pd
import numpy as np

import itertools

# import multiprocessing
# from multiprocessing import Process
# from joblib import Parallel, delayed

#nearest_next,
# nearest_previous)(nearest, nearest_nonoverlapping,

from collections import OrderedDict

from sorted_nearest import (nearest_previous_nonoverlapping,
                            nearest_next_nonoverlapping,
                            nearest_nonoverlapping, find_clusters)
try:
    dummy = profile
except:
    profile = lambda x: x


# from clustertree import find_clusters

import pyranges as pr

from time import time
import datetime

from collections import defaultdict

def pick_out_indexes_possibly_nonunique(df, indexes, invert=False):

    if isinstance(indexes, list) and indexes:
        concat = np.concatenate(indexes)
        indexes = np.unique(concat)

    if not invert:
        return df.loc[df.index.isin(indexes)]
    else:
        return df.loc[~df.index.isin(indexes)]


def _overlap(self, other, strandedness, invert, how=None):

    assert strandedness in ["same", "opposite", False, None]

    df = self.df
    other_strand = {"+": "-", "-": "+"}
    same_strand = {"+": "+", "-": "-"}

    if strandedness == "same":
        strand_dict = same_strand
    elif strandedness == "opposite":
        strand_dict = other_strand

    dfs = OrderedDict()

    if self.stranded and other.stranded and strandedness:

        indexes_of_overlapping = []
        for (chromosome, strand), cdf in df.groupby("Chromosome Strand".split()):

            starts = cdf.Start.values
            ends = cdf.End.values
            indexes = cdf.index.values

            strand = strand_dict[strand]
            it = other.__ncls__[chromosome, strand]
            if how == "containment":
                indexes = it.has_containment(starts, ends, indexes)
            else:
                indexes = it.has_overlaps(starts, ends, indexes)


            dfs[chromosome, strand] = pick_out_indexes_possibly_nonunique(cdf, indexes, invert)

    else:

        for chromosome, cdf in df.groupby("Chromosome"):

            starts = cdf.Start.values
            ends = cdf.End.values
            indexes = cdf.index.values

            it = other.__ncls__[chromosome, "+"]
            it2 = other.__ncls__[chromosome, "-"]

            if how == "containment":
                idx1 = it.has_containments(starts, ends, indexes)
                idx2 = it2.has_containments(starts, ends, indexes)
            else:
                idx1 = it.has_overlaps(starts, ends, indexes)
                idx2 = it2.has_overlaps(starts, ends, indexes)

            dfs[chromosome] = pick_out_indexes_possibly_nonunique(cdf, [idx1, idx2], invert)

    return pd.concat(dfs.values())


def both_indexes(self, other, strandedness, how=None):

    assert strandedness in ["same", "opposite", False, None]
    assert how in "containment first".split() + [False, None]

    if strandedness:
        assert self.stranded and other.stranded, \
            "Can only do stranded searches when both PyRanges contain strand info"

    df = self.df

    other_strand = {"+": "-", "-": "+"}
    same_strand = {"+": "+", "-": "-"}

    if strandedness == "same":
        strand_dict = same_strand
    elif strandedness == "opposite":
        strand_dict = other_strand

    self_d = OrderedDict()
    other_d = OrderedDict()

    if how == "first" and not strandedness and other.stranded:
        other_dfs = OrderedDict([(k, v) for k, v in other.df.groupby("Chromosome")])

    if strandedness: # know from assertion above that both have strand info

        for (chromosome, strand), cdf in df.groupby("Chromosome Strand".split()):

            starts = cdf.Start.values
            ends = cdf.End.values
            indexes = cdf.index.values

            strand = strand_dict[strand]
            it = other.__ncls__[chromosome, strand]

            if not how:
                _self_indexes, _other_indexes = it.all_overlaps_both(starts, ends, indexes)
            elif how == "containment":
                _self_indexes, _other_indexes = it.all_containments_both(starts, ends, indexes)
            else:
                _self_indexes, _other_indexes = it.first_overlap_both(starts, ends, indexes)


            self_d[chromosome, strand] = _self_indexes
            other_d[chromosome, strand] = _other_indexes

    # if not strandedness, and other df has strand, need to search both ncls
    elif other.stranded:

        # print("elif other stranded " * 10, )
        # print("self\n", self)
        # print("other\n", other)
        for chromosome, cdf in df.groupby("Chromosome"):

            it = other.__ncls__[chromosome, "+"]
            it2 = other.__ncls__[chromosome, "-"]

            starts = cdf.Start.values
            ends = cdf.End.values
            indexes = cdf.index.values

            if not how:
                # print("all overlaps " * 10)
                _self_indexes, _other_indexes = it.all_overlaps_both(starts, ends, indexes)
                _self_indexes2, _other_indexes2 = it2.all_overlaps_both(starts, ends, indexes)
            elif how == "containment":
                # print("all containments " * 10)
                _self_indexes, _other_indexes = it.all_containments_both(starts, ends, indexes)
                _self_indexes2, _other_indexes2 = it2.all_containments_both(starts, ends, indexes)
            else:
                _self_indexes, _other_indexes = it.first_overlap_both(starts, ends, indexes)
                _self_indexes2, _other_indexes2 = it2.first_overlap_both(starts, ends, indexes)

                if chromosome in other_dfs:
                    odf = other_dfs[chromosome]
                else:
                    continue

                if not len(_other_indexes2):
                    pass
                elif not len(_other_indexes):
                    other_indexes = _other_indexes2
                else:
                    s = pd.Series(np.concatenate([_self_indexes, _self_indexes2])).duplicated(keep=False)
                    _dup1, _dup2 = s[:len(_self_indexes)].values, s[len(_self_indexes):].values
                    # print("dup1", _dup1)
                    # print("dup2", _dup2)
                    # print(_other_indexes[_dup1])
                    # print(_other_indexes2[_dup2])
                    cond = odf.loc[_other_indexes[_dup1]].Start.values < odf.loc[_other_indexes2[_dup2]].Start.values
                    # print("os", odf.loc[_other_indexes[_dup1]].Start.values)
                    # print("os2", odf.loc[_other_indexes2[_dup2]].Start.values)
                    # print("cond", cond)

                    _other_indexes_tmp = np.where(cond, _other_indexes[_dup1], _other_indexes2[_dup2])
                    _self_indexes_tmp = np.where(cond, _self_indexes[_dup1], _self_indexes2[_dup2])
                    # print(_self_indexes, ~_dup1)
                    # print("_self_indexes[~_dup1]", _self_indexes[~_dup1])
                    # print("_self_indexes[~_dup2]", _self_indexes2[~_dup2])
                    # print("_other_indexes[~_dup1]", _other_indexes[~_dup1])
                    # print("_other_indexes[~_dup2]", _other_indexes2[~_dup2])
                    # _self_indexes2 = np.concatenate(, _self_indexes2[~_dup2])
                    _self_indexes2 = np.concatenate([_self_indexes[~_dup1], _self_indexes2[~_dup2]])
                    _other_indexes2 = np.concatenate([_other_indexes[~_dup1], _other_indexes2[~_dup2]])

                    _self_indexes = _self_indexes_tmp
                    _other_indexes = _other_indexes_tmp
                    # print("~cond", ~cond)
                    # print("~_dup1", ~_dup1, _dup1)
                    # print("~_dup2", ~_dup2, _dup2)
                    # print("~_dup1", _other_indexes[~_dup1], _self_indexes[~_dup1])
                    # print("~_dup2", _other_indexes[~_dup2], _self_indexes[~_dup2])
                    # _other_indexes = np.where(~cond, _other_indexes[~_dup1], _other_indexes2[~_dup2])
                    # _self_indexes = np.where(~cond, _self_indexes[~_dup1], _self_indexes2[~_dup2])

            self_d[chromosome] = np.concatenate([_self_indexes, _self_indexes2])
            other_d[chromosome] = np.concatenate([_other_indexes, _other_indexes2])

            if not len(_self_indexes) or not len(_self_indexes2):
                self_d[chromosome] = np.array(self_d[chromosome], dtype=int)
                other_d[chromosome] = np.array(other_d[chromosome], dtype=int)


    # if not strandedness, and other df has no strand info, need to search one ncls
    elif not other.stranded:

        for chromosome, cdf in df.groupby("Chromosome"):

            it = other.__ncls__[chromosome]

            starts = cdf.Start.values
            ends = cdf.End.values
            indexes = cdf.index.values

            if not how:
                _self_indexes, _other_indexes = it.all_overlaps_both(starts, ends, indexes)
            elif how == "containment":
                _self_indexes, _other_indexes = it.all_containments_both(starts, ends, indexes)
            else:
                _self_indexes, _other_indexes = it.first_overlap_both(starts, ends, indexes)

            self_d[chromosome] = _self_indexes
            other_d[chromosome] = _other_indexes

    return self_d, other_d


def sort_one_by_one(d, col1, col2):
    """
    Equivalent to pd.sort_values(by=[col1, col2]), but faster.
    """
    d = d.sort_values(by=[col2])
    return d.sort_values(by=[col1], kind='mergesort')


def _cluster(self, strand=False, maxdist=0, minnb=1):

    dfs = []

    idx_start, idx_end = 0, 0
    if strand:

        for (c, s), cdf in self.df.groupby(["Chromosome", "Strand"]):
            cdf = cdf.sort_values("Start")
            starts, ends = find_clusters(cdf.Start.values, cdf.End.values)
            df = pd.DataFrame({"Chromosome": c, "Start": starts, "End": ends, "Strand": s})
            dfs.append(df)

        df = pd.concat(dfs, ignore_index=True)["Chromosome Start End Strand".split()]

    else:

        for c, cdf in self.df.groupby(["Chromosome"]):
            cdf = cdf.sort_values("Start")
            starts, ends = find_clusters(cdf.Start.values, cdf.End.values)
            df = pd.DataFrame({"Chromosome": c, "Start": starts, "End": ends})
            dfs.append(df)

        df = pd.concat(dfs, ignore_index=True)["Chromosome Start End".split()]


    return df



def _tile(self, tile_size=50):

    "No no no this is slow! Write in C."

    df = self.df.copy()

    df.Start = df.Start - (df.Start % tile_size)
    df.End = df.End - (df.End % tile_size) + tile_size - 1

    rows = []
    for _, row in df.iterrows():
        d = row.to_dict()
        for tile in range(row.Start, row.End, tile_size):
            d2 = d.copy()
            d2["Start"], d2["End"] = tile, tile + tile_size - 1
            rows.append(d2)

    df = pd.DataFrame.from_dict(rows)[df.columns]

    return df


def invert(self):

    pass


# def _multithreaded_intersection(key, scdf, ocdf, strandedness=None, how=None):


#     # is there anything to gain?

#     # create ncls

#     # both indexes

#     # intersect op

#     # c, s = list(ocdf.head(0)["Chromosome Strand".split()].values)

#     starts = np.copy(scdf.Start.values)
#     ends = np.copy(scdf.End.values)
#     indexes = np.copy(scdf.index.values)

#     it = pr.PyRanges(ocdf).__ncls__[key]

#     _self_indexes, _other_indexes = it.all_overlaps_both(starts, ends, indexes)

#     scdf = scdf.loc[_self_indexes]
#     ocdf = ocdf.loc[_other_indexes, ["Start", "End"]]

#     new_starts = pd.Series(
#         np.where(scdf.Start.values > ocdf.Start.values, scdf.Start, ocdf.Start),
#         index=scdf.index, dtype=np.long)

#     new_ends = pd.Series(
#         np.where(scdf.End.values < ocdf.End.values, scdf.End, ocdf.End),
#         index=scdf.index, dtype=np.long)

#     pd.options.mode.chained_assignment = None  # default='warn'
#     scdf.loc[:, "Start"] = new_starts
#     scdf.loc[:, "End"] = new_ends
#     pd.options.mode.chained_assignment = 'warn'

#     return scdf




def _intersection(self, other, strandedness=None, how=None):

    sidx, oidx = both_indexes(self, other, strandedness, how)

    other_strand = {"+": "-", "-": "+"}

    if len(list(sidx.keys())[0]) == 2: # chromosome and strand
        grpby_key = "Chromosome Strand".split()
    else:
        grpby_key = "Chromosome"

    other_dfs = {k: d for k, d in other.df.groupby(grpby_key)}

    dfs = []

    for key, scdf in self.df.groupby(grpby_key):

        if len(key) == 2 and strandedness == "opposite":
            key = key[0], other_strand[key[1]]

        if not key in other_dfs:
            continue

        ocdf = other_dfs[key]

        scidx = sidx[key]
        ocidx = oidx[key]

        scdf = scdf.loc[scidx]
        ocdf = ocdf.loc[ocidx, ["Start", "End"]]

        new_starts = pd.Series(
            np.where(scdf.Start.values > ocdf.Start.values, scdf.Start, ocdf.Start),
            index=scdf.index, dtype=np.long)

        new_ends = pd.Series(
            np.where(scdf.End.values < ocdf.End.values, scdf.End, ocdf.End),
            index=scdf.index, dtype=np.long)

        pd.options.mode.chained_assignment = None  # default='warn'
        scdf.loc[:, "Start"] = new_starts
        scdf.loc[:, "End"] = new_ends
        pd.options.mode.chained_assignment = 'warn'

        dfs.append(scdf)

    if dfs:
        df = pd.concat(dfs)
    else:
        df = pd.DataFrame(columns="Chromosome Start End Strand".split())

    return df



def _coverage(ranges, value_col=None, stranded=False):

    try:
        from pyranges import PyRles
    except ImportError:
        raise Exception("Using the coverage method requires that pyrle is installed.")

    return PyRles(ranges, value_col=value_col, stranded=stranded)


def _keep_both(idx_df, other, suffixes, new_pos):

    self_suffix, other_suffix = suffixes

    if new_pos:
        assert self_suffix and other_suffix, "If new_pos is union or intersection, you must give two suffixes of length > 0."
    else:
        assert other_suffix, "Must give a suffix for the right dataframe."

    if not new_pos:
        suffixes = ["", other_suffix]
        to_drop = ["IndexOther", "Chromosome" + other_suffix]
        self_suffix = ""
        to_rename = {}
    else:
        suffixes = [self_suffix, other_suffix]
        to_drop = ["IndexOther", "Chromosome" + other_suffix]
        to_rename = {"Chromosome" + self_suffix: "Chromosome"}

    idx_df = idx_df.merge(other.df, left_on="IndexOther", right_index=True, suffixes=suffixes)
    idx_df = idx_df.drop(to_drop, axis=1)
    idx_df = idx_df.rename(to_rename, axis="columns")

    if new_pos == "intersect":
        new_starts = np.where(idx_df["Start" + self_suffix] > idx_df["Start" + other_suffix], idx_df["Start" + self_suffix], idx_df["Start" + other_suffix])
        new_ends = np.where(idx_df["End" + self_suffix] < idx_df["End" + other_suffix], idx_df["End" + self_suffix], idx_df["End" + other_suffix])
        idx_df.insert(1, "Start", new_starts)
        idx_df.insert(2, "End", new_ends)

    if new_pos == "union":
        new_starts = np.where(idx_df["Start" + self_suffix] < idx_df["Start" + other_suffix], idx_df["Start" + self_suffix], idx_df["Start" + other_suffix])
        new_ends = np.where(idx_df["End" + self_suffix] > idx_df["End" + other_suffix], idx_df["End" + self_suffix], idx_df["End" + other_suffix])
        idx_df.insert(1, "Start", new_starts)
        idx_df.insert(2, "End", new_ends)

    return idx_df


def _overlap_write_both(self, other, strandedness=False, new_pos=None, suffixes=["_a", "_b"], suffix="_b", how=None):

    assert new_pos in ["intersection", "union", False, None]

    sidx, oidx = both_indexes(self, other, strandedness, how)

    other_strand = {"+": "-", "-": "+"}

    if len(list(sidx.keys())[0]) == 2: # chromosome and strand
        grpby_key = "Chromosome Strand".split()
    else:
        grpby_key = "Chromosome"

    other_dfs = {k: d for k, d in other.df.groupby(grpby_key)}

    dfs = []

    idx_counter, current_end = 0, 0
    for key, scdf in self.df.groupby(grpby_key):

        if len(key) == 2 and strandedness == "opposite":
            key = key[0], other_strand[key[1]]

        if not key in other_dfs:
            continue

        ocdf = other_dfs[key].drop("Chromosome", 1)

        scidx = sidx[key]
        ocidx = oidx[key]

        scdf = scdf.loc[scidx]
        ocdf = ocdf.loc[ocidx]

        pd.options.mode.chained_assignment = None  # default='warn'
        pd.options.mode.chained_assignment = 'warn'

        current_end += len(scdf)
        scdf.index = range(idx_counter, current_end)
        ocdf.index = range(idx_counter, current_end)
        idx_counter += len(scdf)

        if not new_pos:
            _df = scdf.join(ocdf, rsuffix=suffix)

        elif new_pos == "intersection":

            new_starts = pd.Series(
                np.where(scdf.Start.values > ocdf.Start.values, scdf.Start, ocdf.Start),
                index=scdf.index, dtype=np.long)

            new_ends = pd.Series(
                np.where(scdf.End.values < ocdf.End.values, scdf.End, ocdf.End),
                index=scdf.index, dtype=np.long)
            _df = scdf.join(ocdf, lsuffix=suffixes[0], rsuffix=suffixes[1])
            _df.insert(1, "Start", new_starts)
            _df.insert(2, "End", new_ends)
            _df.rename(index=str, columns={"Chromosome" + suffixes[0]: "Chromosome", "Strand" + suffixes[0]: "Strand"}, inplace=True)

        elif new_pos == "union":

            new_starts = pd.Series(
                np.where(scdf.Start.values < ocdf.Start.values, scdf.Start, ocdf.Start),
                index=scdf.index, dtype=np.long)

            new_ends = pd.Series(
                np.where(scdf.End.values > ocdf.End.values, scdf.End, ocdf.End),
                index=scdf.index, dtype=np.long)
            _df = scdf.join(ocdf, lsuffix=suffixes[0], rsuffix=suffixes[1])
            _df.insert(1, "Start", new_starts)
            _df.insert(2, "End", new_ends)
            _df.rename(index=str, columns={"Chromosome" + suffixes[0]: "Chromosome", "Strand" + suffixes[0]: "Strand"}, inplace=True)

        dfs.append(_df)


    if dfs:
        df = pd.concat(dfs)
    else:
        return pd.DataFrame(columns="Chromosome Start End Strand".split())

    return df


def _set_intersection(self, other, strandedness=None, how=None):

    strand = True if strandedness else False
    s = self.cluster(strand=strand)
    o = other.cluster(strand=strand)

    return _intersection(s, o, strandedness, how)


def _set_union(self, other, strand):

    if strand:
        assert self.stranded and other.stranded, \
            "Can only do stranded set union when both PyRanges contain strand info."

    if len(self) == 0:
        return _cluster(other, strand=strand)
    elif len(other) == 0:
        return _cluster(self, strand=strand)

    dfs = []

    if strand and len(list(self.__ncls__.keys())[0]) == 2: # chromosome and strand
        grpby_key = "Chromosome Strand".split()
        columns = "Chromosome Start End Strand".split()
    else:
        grpby_key = "Chromosome"
        columns = "Chromosome Start End".split()

    self_dfs =  {k: d for k, d in self.df.groupby(grpby_key)}
    other_dfs = {k: d for k, d in other.df.groupby(grpby_key)}

    idx_start, idx_end = 0, 0
    for key in set(self_dfs).union(other_dfs):

        if key in other_dfs and key in self_dfs:
            _starts = np.concatenate([
                self_dfs[key].Start.values,
                other_dfs[key].Start.values])
            _ends = np.concatenate([
                self_dfs[key].End.values,
                other_dfs[key].End.values])
        elif key in self_dfs and not key in other_dfs:
            _starts = self_dfs[key].Start.values
            _ends = self_dfs[key].End.values
        elif key in other_dfs and not key in self_dfs:
            _starts = other_dfs[key].Start.values
            _ends = other_dfs[key].End.values


        cdf = pd.DataFrame({"Start": _starts, "End": _ends})["Start End".split()]
        # print("cdf" * 10 + "\n", cdf)
        # cdf = sort_one_by_one(cdf, "Start", "End")
        cdf = cdf.sort_values("Start")
        # print("cdf" * 10 + "\n", cdf)
        # print(clusters)
        starts, ends = find_clusters(cdf.Start.values, cdf.End.values)
        # print(starts, ends)
        idx_end += len(starts)

        if strand:
            _df = pd.DataFrame({"Chromosome": key[0], "Start": starts, "End": ends, "Strand": key[1]})
        else:
            _df = pd.DataFrame({"Chromosome": key, "Start": starts, "End": ends})

        _df.index = range(idx_start, idx_end)
        dfs.append(_df)
        idx_start = idx_end

    df = pd.concat(dfs)[columns]

    return df


def _subtraction(self, other, strandedness):

    #print(self.df.to_csv(sep=" "))
    strand = False
    if strandedness:
        assert "Strand" in self.df and "Strand" in other.df, \
            "Can only do stranded set difference when both PyRanges contain strand info."
        strand = True

    if len(self) == 0:
        return pd.DataFrame(columns="Chromosome Start End Strand".split())
    if len(other) == 0:
        return self.df

    other = other.cluster(strand=strand)

    other_strand = {"+": "-", "-": "+"}
    same_strand = {"+": "+", "-": "-"}

    if strandedness == "same":
        strand_dict = same_strand
    elif strandedness == "opposite":
        strand_dict = other_strand

    if strandedness and self.stranded: # chromosome and strand
        grpby_key = "Chromosome Strand".split()
    else:
        grpby_key = "Chromosome"

    # print("self.df", self.df.to_csv(sep=" "))


    self_dfs = {k: d for k, d in self.df.groupby(grpby_key)}
    other_dfs = {k: d for k, d in other.df.groupby(grpby_key)}

    dfs = []

    # need to store indexes in dict since if it is opposite stranded
    # cannot do removal in same iteration
    _idxes = dict()

    for key, scdf in self_dfs.items():

        # the downstream union of self/missing indexes sorts them
        # therefore, we need them to be in correct order here
        scdf.index = pd.Index(range(len(scdf)))
        print("-------")
        print("scdf", scdf.to_csv(sep=" "))


        if len(key) == 2:
            old_key = key
            key = key[0], strand_dict[key[1]]


        if not key in other_dfs:
            dfs.append(scdf)
            continue

        # scdf = self_dfs[key]
        ogr = pr.PyRanges(other_dfs[key])
        # ogr.index = pd.Index(range(len(ogr)))
        #print("scdf\n", scdf)
        #print("odf\n", ogr)

        # check which self reads have partial overlap
        if not ogr.stranded and not strandedness:
            #print("1" * 100)
            idx_self, new_starts, new_ends = ogr.__ncls__[key].set_difference_helper(
               scdf.Start.values,
                scdf.End.values,
                scdf.index.values)

        elif self.stranded and ogr.stranded and strandedness:

            idx_self, new_starts, new_ends = ogr.__ncls__[key].set_difference_helper(
                scdf.Start.values,
                scdf.End.values,
                scdf.index.values)

        elif ogr.stranded and not strandedness:

            #print("3" * 100)
            idx_self1, new_starts1, new_ends1 = ogr.__ncls__[key, "+"].set_difference_helper(
                scdf.Start.values,
                scdf.End.values,
                scdf.index.values)
            idx_self2, new_starts2, new_ends2 = ogr.__ncls__[key, "-"].set_difference_helper(
                scdf.Start.values,
                scdf.End.values,
                scdf.index.values)

            idx_self = np.concatenate([idx_self1, idx_self2])

            if not idx_self.dtype == np.int64:
                idx_self = np.array(idx_self, dtype=np.int64)

            new_starts = np.concatenate([new_starts1, new_starts2])
            new_ends = np.concatenate([new_ends1, new_ends2])

        #print("scdf.index", scdf.index)
        missing_idx = pd.Index(scdf.index).difference(idx_self)
        #print("missing_idx " * 10, missing_idx)
        idx_to_drop = new_starts != -1
        #print(idx_to_drop)
        new_starts = new_starts[idx_to_drop]
        new_ends = new_ends[idx_to_drop]

        idx_self = idx_self[idx_to_drop]
        new_starts = pd.Series(new_starts, index=idx_self).sort_index()
        new_ends = pd.Series(new_ends, index=idx_self).sort_index()
        # idx_self = np.sort(idx_self)

        print("scdf", scdf.to_csv(sep=" "))
        print("scdf.index", scdf.index)
        print("idx_self", idx_self)
        print("missing_idx", missing_idx)
        scdf = scdf.reindex(missing_idx.union(idx_self))
        print("scdf", scdf.to_csv(sep=" "))
        print("scdf.index", scdf.index)
        print("idx_self", idx_self)

        if len(idx_self):
            # why does boolean indexing work, but not regular?

            # pd.options.mode.chained_assignment = None  # default='warn'
            #print("idx_self", idx_self)
            scdf.loc[scdf.index.isin(idx_self), "Start"] = new_starts
            scdf.loc[scdf.index.isin(idx_self), "End"] = new_ends

        # find those in scdf that were not in idx_self

        #print("scdf", scdf)

        dfs.append(scdf)

    if dfs:
        df = pd.concat(dfs)
    else:
        return pd.DataFrame(columns="Chromosome Start End Strand".split())

    #print("dfs", dfs)
    #print("df", df)
    return df


# def _next_nonoverlapping(left_ends, right_starts):

#     left_ends = left_ends.sort_values()
#     right_starts = right_starts.sort_values()
#     l_idx, r_idx, dist = nearest_next_nonoverlapping(left_ends.values - 1, left_ends.index.values, right_starts.values, right_starts.index.values)

#     return l_idx, r_idx, dist


# def _previous_nonoverlapping(left_starts, right_ends):

#     left_starts = left_starts.sort_values()
#     right_ends = right_ends.sort_values()
#     l_idx, r_idx, dist = nearest_previous_nonoverlapping(left_starts.values, left_starts.index.values, right_ends.values - 1, right_ends.index.values)

#     return l_idx, r_idx, dist

def _next_nonoverlapping(left_ends, right_starts, right_indexes):

    left_ends = left_ends.sort_values()
    right_starts = right_starts.sort_values()
    r_idx, dist = nearest_next_nonoverlapping(left_ends.values - 1, right_starts.values, right_indexes)
    r_idx = pd.Series(r_idx, index=left_ends.index).sort_index().values
    dist = pd.Series(dist, index=left_ends.index).sort_index().values

    return r_idx, dist


def _previous_nonoverlapping(left_starts, right_ends, right_indexes):

    left_starts = left_starts.sort_values()
    right_ends = right_ends.sort_values()
    r_idx, dist = nearest_previous_nonoverlapping(left_starts.values, right_ends.values - 1, right_ends.index.values)
    # print("ridx before", r_idx)
    r_idx = pd.Series(r_idx, index=left_starts.index).sort_index().values
    dist = pd.Series(dist, index=left_starts.index).sort_index().values
    # print("ridx after", r_idx)

    return r_idx, dist

# from memory_profiler import profile


def _overlapping_for_nearest(self, other, strandedness, suffix):
    nearest_df = pd.DataFrame(columns="Chromosome Start End Strand".split())

    sidx, oidx = both_indexes(self, other, strandedness, how="first")

    if list(oidx.values()):
        idxs = np.concatenate(list(sidx.values()))
        missing_overlap = self.df.index[~self.df.index.isin(idxs)]
        # print("missing overlap", missing_overlap)
        df_to_find_nearest_in = self.df.reindex(missing_overlap)

        odf = other.df.reindex(np.concatenate(list(oidx.values())))
        odf.index = np.concatenate(list(sidx.values()))
        sdf = self.df.reindex(np.concatenate(list(sidx.values())))

        nearest_df = sdf.join(odf, rsuffix=suffix).drop("Chromosome" + suffix, axis=1)
        nearest_df.insert(nearest_df.shape[1], "Distance", 0)
    else:
        df_to_find_nearest_in = self.df

    return nearest_df, df_to_find_nearest_in


def _nearest(self, other, strandedness, suffix="_b", how=None, overlap=True):

    if overlap:
        nearest_df, df_to_find_nearest_in = _overlapping_for_nearest(self, other, strandedness, suffix)
    else:
        df_to_find_nearest_in = self.df

    other_strand = {"+": "-", "-": "+"}

    if self.stranded and strandedness: # chromosome and strand
        grpby_key = "Chromosome Strand".split()
    else:
        grpby_key = "Chromosome"

    other_dfs = {k: d for k, d in other.df.groupby(grpby_key)}

    dfs = []

    for key, scdf in df_to_find_nearest_in.groupby(grpby_key):

        if len(key) == 2 and strandedness == "opposite":
            other_key = key[0], other_strand[key[1]]
        else:
            other_key = key

        if not other_key in other_dfs:
            continue

        ocdf = other_dfs[other_key]

        scdf.index = pd.Index(range(len(scdf)))

        if how == "next":
            r_idx, dist = _next_nonoverlapping(scdf.End, ocdf.Start, ocdf.index.values)
        elif how == "previous":
            r_idx, dist = _previous_nonoverlapping(scdf.Start, ocdf.End, ocdf.index.values)
        else:
            previous_r_idx, previous_dist = _previous_nonoverlapping(scdf.Start, ocdf.End, ocdf.index.values)

            next_r_idx, next_dist = _next_nonoverlapping(scdf.End, ocdf.Start, ocdf.index.values)

            r_idx, dist = nearest_nonoverlapping(previous_r_idx,
                                                 previous_dist,
                                                 next_r_idx, next_dist)

        ocdf = ocdf.reindex(r_idx, fill_value=-1) # instead of np.nan, so ints are not promoted to float

        ocdf.index = scdf.index
        ocdf.insert(ocdf.shape[1], "Distance", pd.Series(dist, index=ocdf.index).fillna(-1).astype(int))
        ocdf.drop("Chromosome", axis=1, inplace=True)

        r_idx = pd.Series(r_idx, index=ocdf.index)
        scdf = scdf.drop(r_idx.loc[r_idx == -1].index)

        result = scdf.join(ocdf, rsuffix=suffix)

        dfs.append(result)

    if dfs:
        df = pd.concat(dfs)
    else:
        df = pd.DataFrame(columns="Chromosome Start End Strand".split())


    if overlap and not df.empty and not nearest_df.empty:
        df = pd.concat([nearest_df, df])
    elif overlap and not nearest_df.empty:
        df = nearest_df

    return df


def _slack(self, slack):

    df = self.df.copy()
    df.Start = df.Start - slack
    df.loc[df.Start < 0, "Start"] = 0
    df.End = df.End + slack

    return df


def _tss(self, slack=0):

    df = self.df

    tss_pos = df.loc[df.Strand == "+"]

    tss_neg = df.loc[df.Strand == "-"].copy()

    # pd.options.mode.chained_assignment = None
    tss_neg.loc[:, "Start"] = tss_neg.End

    # pd.options.mode.chained_assignment = "warn"
    tss = pd.concat([tss_pos, tss_neg], sort=False)
    tss["End"] = tss.Start
    tss.End = tss.End + 1 + slack
    tss.Start = tss.Start - slack
    tss.loc[tss.Start < 0, "Start"] = 0

    return tss.reindex(df.index)


def _tes(self, slack=0):

    df = self.df

    tes_pos = df.loc[df.Strand == "+"]

    tes_neg = df.loc[df.Strand == "-"].copy()

    # pd.options.mode.chained_assignment = None
    tes_neg.loc[:, "Start"] = tes_neg.End

    # pd.options.mode.chained_assignment = "warn"
    tes = pd.concat([tes_pos, tes_neg], sort=False)
    tes["Start"] = tes.End
    tes.End = tes.End + 1 + slack
    tes.Start = tes.Start - slack
    tes.loc[tes.Start < 0, "Start"] = 0

    return tes.reindex(df.index)


def _lengths(self):

    df = self.df
    lengths = df.End - df.Start

    return lengths


def _jaccard(self, other, strandedness):

    if strandedness:
        strand = True
    else:
        strand = False

    il = self.set_intersection(other, strandedness).lengths().sum()
    # ul = self.set_union(other, strandedness).lengths().sum()
    s = self.cluster(True).lengths().sum()
    o = other.cluster(True).lengths().sum()

    # print("o", o)
    # print("s", s)
    # print("il", il)
    # print("ul", ul)



    # if il == ul:
    #     return 1
    if s + o == il:
        return 1

    return il / (s + o - il)
